/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

#include <iostream>
//Added by qt3to4:
#include <QPixmap>
#include <QMouseEvent>
#include <QEvent>
#include <Q3PointArray>
#include <Q3PopupMenu>
#include <QDialogButtonBox>
#include <QSpinBox>
using std::cerr; using std::endl;
/*
  Quarticurve style.
  Copyright (c) 2002 Red Hat, Inc.
  Copyright (c) 2007-2008 Kevin Kofler
  Authors: Bernhard Rosenkränzer <bero@redhat.com>
           Preston Brown <pbrown@redhat.com>
           Than Ngo <than@redhat.com>
           Alexander Larsson <alexl@redhat.com>
           Chris Lee <clee@redhat.com>
  Ported to Qt 4 by Kevin Kofler <Kevin@tigcc.ticalc.org>

  Released under the GNU General Public License (GPL) v2.

  Based on the KDE Light style, 2nd revision:
  Copyright (c) 2000-2001 Trolltech AS (info@trolltech.com)

  Permission is hereby granted, free of charge, to any person obtaining a
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation
  the rights to use, copy, modify, merge, publish, distribute, sublicense,
  and/or sell copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
*/

#include ".moc/moc_quarticurve.cpp"

#include "qmenubar.h"
#include "qapplication.h"
#include "qpainter.h"
#include "qpalette.h"
#include "q3frame.h"
#include "qpushbutton.h"
#include "qdrawutil.h"
#include "qscrollbar.h"
#include "qtabbar.h"
#include "qpointer.h"
#include "qlayout.h"
#include "qlineedit.h"
#include "q3listview.h"
#include "qbitmap.h"
#include "qimage.h"
#include "qcombobox.h"
#include "qslider.h"
#include "qstylefactory.h"
#include "q3cleanuphandler.h"
#include "qcheckbox.h"
#include "qradiobutton.h"
#include "q3progressbar.h"
#include "qcursor.h"

#define MAX(x, y) ((x) > (y) ? (x) : (y))
#define MIN(x, y) ((x) < (y) ? (x) : (y))

#define RADIO_SIZE 13
#define CHECK_SIZE 13

#include "bits.h"

const double QuarticurveStyle::shadeFactors[8] = {1.065, 0.963, 0.896, 0.85, 0.768, 0.665, 0.4, 0.205};

static void rgb_to_hls (double *r, double *g, double *b) {
	double min;
	double max;
	double red;
	double green;
	double blue;
	double h, l, s;
	double delta;

	red = *r;
	green = *g;
	blue = *b;

	if (red > green) {
		if (red > blue)
			max = red;
		else
			max = blue;

		if (green < blue)
			min = green;
		else
			min = blue;
	} else {
		if (green > blue)
			max = green;
		else
			max = blue;

		if (red < blue)
			min = red;
		else
			min = blue;
	}

	l = (max + min) / 2;
	s = 0;
	h = 0;

	if (max != min) {
		if (l <= 0.5)
			s = (max - min) / (max + min);
		else
			s = (max - min) / (2 - max - min);

		delta = max -min;
		if (red == max)
			h = (green - blue) / delta;
		else if (green == max)
			h = 2 + (blue - red) / delta;
		else if (blue == max)
			h = 4 + (red - green) / delta;

		h *= 60;
		if (h < 0.0)
			h += 360;
	}

	*r = h;
	*g = l;
	*b = s;
}

static void hls_to_rgb (double *h, double *l, double *s) {
	double hue;
	double lightness;
	double saturation;
	double m1, m2;
	double r, g, b;

	lightness = *l;
	saturation = *s;

	if (lightness <= 0.5)
		m2 = lightness * (1 + saturation);
	else
		m2 = lightness + saturation - lightness * saturation;
	m1 = 2 * lightness - m2;

	if (saturation == 0) {
		*h = lightness;
		*l = lightness;
		*s = lightness;
	} else {
		hue = *h + 120;
		while (hue > 360)
			hue -= 360;
		while (hue < 0)
			hue += 360;

		if (hue < 60)
			r = m1 + (m2 - m1) * hue / 60;
		else if (hue < 180)
			r = m2;
		else if (hue < 240)
			r = m1 + (m2 - m1) * (240 - hue) / 60;
		else
			r = m1;

		hue = *h;
		while (hue > 360)
			hue -= 360;
		while (hue < 0)
			hue += 360;

		if (hue < 60)
			g = m1 + (m2 - m1) * hue / 60;
		else if (hue < 180)
			g = m2;
		else if (hue < 240)
			g = m1 + (m2 - m1) * (240 - hue) / 60;
		else
			g = m1;

		hue = *h - 120;
		while (hue > 360)
			hue -= 360;
		while (hue < 0)
			hue += 360;

		if (hue < 60)
			b = m1 + (m2 - m1) * hue / 60;
		else if (hue < 180)
			b = m2;
		else if (hue < 240)
			b = m1 + (m2 - m1) * (240 - hue) / 60;
		else
			b = m1;

		*h = r;
		*l = g;
		*s = b;
	}
}

static void shade (const QColor &ca, QColor &cb, double k) {
	int r,g,b;
	double red;
	double green;
	double blue;

	ca.rgb (&r, &g, &b);

	red = (double) r / 255.0;
	green = (double) g / 255.0;
	blue = (double) b / 255.0;

	rgb_to_hls (&red, &green, &blue);

	green *= k;
	if (green > 1.0)
		green = 1.0;
	else if (green < 0.0)
		green = 0.0;

	blue *= k;
	if (blue > 1.0)
		blue = 1.0;
	else if (blue < 0.0)
		blue = 0.0;

	hls_to_rgb (&red, &green, &blue);

	cb.setRgb ((int)(red*255.0), (int)(green*255.0), (int)(blue*255.0));
}

static QImage *
generate_bit (unsigned char alpha[], const QColor &color, double mult)
{
  unsigned int r, g, b;
  QImage *image;
  QRgb *pixels;
  int w, h;
  int x, y;
  
  r = (int) (color.red() * mult);
  r = MIN(r, 255);
  g = (int) (color.green() * mult);
  g = MIN(g, 255);
  b = (int) (color.blue() * mult);
  b = MIN(b, 255);
  
  image = new QImage (RADIO_SIZE, RADIO_SIZE, 32);
  image->setAlphaBuffer(true);

  w = image->width();
  h = image->height();

  for (y=0; y < h; y++)
    {
      pixels = (QRgb *)image->scanLine(y);
      for (x=0; x < w; x++)
        pixels[x] = qRgba (r, g, b, alpha?alpha[y*w+x]:255);
    }

  return image;
}

#define CLAMP(x, low, high)  (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))
#define CLAMP_UCHAR(v) ((unsigned char) (CLAMP (((int)v), (int)0, (int)255)))

static QImage *
colorize_bit (unsigned char *bit,
              unsigned char *alpha,
              const QColor  &new_color)
{
  QImage *image;
  double intensity;
  int x, y;
  const unsigned char *src, *asrc;
  QRgb *dest;
  int dest_rowstride;
  int width, height;
  unsigned char *dest_pixels;
  
  image = new QImage (RADIO_SIZE, RADIO_SIZE, 32, 0, QImage::BigEndian);
  image->setAlphaBuffer(true);

  if (image == NULL)
    return NULL;
  
  dest_rowstride = image->bytesPerLine();
  width = image->width();
  height = image->height();
  dest_pixels = image->bits();
  
  for (y = 0; y < RADIO_SIZE; y++)
    {
      src = bit + y * RADIO_SIZE;
      asrc = alpha + y * RADIO_SIZE;
      dest = (QRgb *)image->scanLine (y);

      for (x = 0; x < RADIO_SIZE; x++)
        {
          int dr, dg, db;
          
          intensity = src[x]/ 255.0;

          if (intensity <= 0.5)
            {
              /* Go from black at intensity = 0.0 to new_color at intensity = 0.5 */
              dr = int((new_color.red() * intensity * 2.0));
              dg = int((new_color.green() * intensity * 2.0));
              db = int((new_color.blue() * intensity * 2.0));
            }
          else
            {
              /* Go from new_color at intensity = 0.5 to white at intensity = 1.0 */
              dr = int((new_color.red() + (255 - new_color.red()) * (intensity - 0.5) * 2.0));
              dg = int((new_color.green() + (255 - new_color.green()) * (intensity - 0.5) * 2.0));
              db = int((new_color.blue() + (255 - new_color.blue()) * (intensity - 0.5) * 2.0));
            }

          dest[x] = qRgba (CLAMP_UCHAR (dr), CLAMP_UCHAR (dg), CLAMP_UCHAR (db), asrc[x]);
        }
    }

  return image;
}


static void
composeImage (QImage *destImg, QImage *srcImg)
{
  int w, h, x, y;
  QRgb *src, *dest;
  unsigned int a;
  QRgb s, d;

  w = destImg->width();
  h = destImg->height();

  for (y = 0; y < h; y++)
    {
      src = (QRgb *)srcImg->scanLine(y);
      dest = (QRgb *)destImg->scanLine(y);

      for (x = 0; x < w; x++)
        {
          s = src[x];
          d = dest[x];
          
          a = qAlpha(s);

          dest[x] = qRgba ((qRed(s) * a + (255-a)*qRed(d)) / 255,
                           (qGreen(s) * a + (255-a)*qGreen(d)) / 255,
                           (qBlue(s) * a + (255-a)*qBlue(d)) / 255,
                           a + ((255-a)*qAlpha(d)) / 255);
        }
    }
}

struct QuarticurveStylePrivate
{
    QuarticurveStylePrivate()
        : hoverWidget(0), hovering(FALSE), sliderActive(FALSE), mousePressed(FALSE),
          scrollbarElement(0), lastElement(0), ref(1)
    { ; }

    QPointer<QWidget> hoverWidget;
    bool hovering, sliderActive, mousePressed;
    int scrollbarElement, lastElement, ref;
    QPoint mousePos;
};

static QuarticurveStylePrivate * singleton = 0;

/* We assume this seldom collides, since we can only cache one at a time */
static long
hashColorGroup (const QColorGroup &cg)
{
  return cg.button().rgb() << 8 ^ cg.highlight().rgb();
}

QuarticurveStyle::QuarticurveColorData::~QuarticurveColorData()
{
  int i;

  for (i = 0; i < 8; i++) {
    if (radioPix[i] != 0)
      delete radioPix[i];
  }
  if (radioMask != 0)
    delete radioMask;

  for (i = 0; i < 6; i++) {
    if (checkPix[i] != 0)
      delete checkPix[i];
  }
}

QuarticurveStyle::QuarticurveStyle() : QCleanlooksStyle(), m_dataCache ()
{
    if ( !singleton )
        singleton = new QuarticurveStylePrivate;
    else
        singleton->ref++;

	m_dataCache.setAutoDelete(true);
}

QuarticurveStyle::~QuarticurveStyle()
{
    if ( singleton && singleton->ref-- <= 0) {
        delete singleton;
        singleton = 0;
    }
}

/* from Qt's motifplus style */
void QuarticurveStyle::polish(QWidget *widget)
{
  if (widget->inherits("QButton") ||
      widget->inherits("QComboBox"))
    widget->installEventFilter(this);

  if (widget->inherits("QScrollBar") ||
      widget->inherits("QSlider")) {
    widget->setMouseTracking(TRUE);
    widget->installEventFilter(this);
  }
  QCleanlooksStyle::polish(widget);
}

void QuarticurveStyle::unpolish(QWidget *widget)
{
    widget->removeEventFilter(this);
    QCleanlooksStyle::unpolish(widget);
}

QuarticurveStyle::QuarticurveColorData *
QuarticurveStyle::realizeData (const QColorGroup &cg) const
{
	QuarticurveColorData *cdata;
	int i, j;

	cdata = new QuarticurveColorData;
	cdata->buttonColor = cg.button().rgb();
	cdata->spotColor = cg.highlight().rgb();

	for (i = 0; i < 8; i++) {
		shade (cg.button(), cdata->shades[i], shadeFactors[i]);
	}

	shade (cg.highlight(), cdata->spots[0], 1.62);
	shade (cg.highlight(), cdata->spots[1], 1.05);
	shade (cg.highlight(), cdata->spots[2], 0.72);

	QImage *dot, *inconsistent, *outline, *circle, *check, *base;

	dot = colorize_bit (dot_intensity, dot_alpha, cg.highlight());
	outline = generate_bit (outline_alpha, cdata->shades[6], 1.0);

	QImage composite (RADIO_SIZE, RADIO_SIZE, 32, 0, QImage::BigEndian);

	for (i = 0; i < 2; i++) {
		for (j = 0; j < 2; j++) {
			if (i == 0) {
				composite.fill (cg.button().rgb());
			} else {
				composite.fill (cg.midlight().rgb());
			}
			composeImage (&composite, outline);

			if (j == 0) {
				circle = generate_bit (circle_alpha, Qt::white, 1.0);
			} else {
				circle = generate_bit (circle_alpha, cdata->shades[1], 1.0);
			}

			composeImage (&composite, circle);
			delete circle;

			cdata->radioPix[i*4+j*2+0] = new QPixmap (composite);

			composeImage (&composite, dot);
			cdata->radioPix[i*4+j*2+1] = new QPixmap (composite);
		}
	}

	QImage mask = outline->createAlphaMask();
	cdata->radioMask = new QPixmap (mask);

	check = generate_bit (check_alpha, cg.highlight(), 1.0);
	inconsistent = generate_bit (check_inconsistent_alpha, cg.highlight(), 1.0);

	for (i = 0; i < 2; i++) {
		if (i == 0) {
			base = generate_bit (check_base_alpha, Qt::white, 1.0);
		} else {
			base = generate_bit (check_base_alpha, cdata->shades[1], 1.0);
		}

		composite.fill (cdata->shades[6].rgb());
		composeImage (&composite, base);
		cdata->checkPix[i*3+0] = new QPixmap (composite);

		composeImage (&composite, check);
		cdata->checkPix[i*3+1] = new QPixmap (composite);

		composite.fill (cdata->shades[6].rgb());
		composeImage (&composite, base);
		composeImage (&composite, inconsistent);
		cdata->checkPix[i*3+2] = new QPixmap (composite);

		delete base;
	}

    // GTK check marks - 0 is highlighted, 1 is normal 
    check = generate_bit (checkmark, cg.highlightedText(), 1.0);
    cdata->checkMark[0] = new QPixmap (*check);
    check = generate_bit (checkmark, cg.buttonText(), 1.0);
    cdata->checkMark[1] = new QPixmap (*check);

	delete dot;
	delete inconsistent;
	delete outline;
	delete check;

	return cdata;
}

const QuarticurveStyle::QuarticurveColorData *
QuarticurveStyle::lookupData (const QColorGroup &cg) const
{
	QuarticurveColorData *cdata;
	long h;
	Q3IntCache<QuarticurveColorData> *cache;

	h = hashColorGroup (cg);

	// Cast away the const crap
	// Nasty, but how else would i do it?
	cache = (Q3IntCache<QuarticurveColorData> *)&m_dataCache;

	cdata = cache->find(h);

	if (cdata == 0 || !cdata->isGroup(cg)) {
		if (cdata != 0) {
			cache->remove (h);
		}

		cdata = realizeData (cg);
		cache->insert (h, cdata);
	}

	return cdata;
}

void
QuarticurveStyle::getShade (const QColorGroup &cg, int shadenr, QColor &res) const
{
	const QuarticurveColorData *cdata = lookupData(cg);

	res = cdata->shades[shadenr];
}

bool QuarticurveStyle::eventFilter(QObject *object, QEvent *event)
{
	switch(event->type()) {
		case QEvent::MouseButtonPress: {
			singleton->mousePressed = TRUE;

			if (! object->inherits("QSlider"))
				break;

			singleton->sliderActive = TRUE;
			break;
		}

		case QEvent::MouseButtonRelease: {
			singleton->mousePressed = FALSE;

			if (! object->inherits("QSlider"))
				break;

			singleton->sliderActive = FALSE;
			((QWidget *) object)->repaint(FALSE);
			break;
		}

		case QEvent::Enter: {
			if (! object->isWidgetType())
				break;

			singleton->hoverWidget = (QWidget *) object;
			if (! singleton->hoverWidget->isEnabled()) {
				singleton->hoverWidget = 0;
				break;
			}
			singleton->hoverWidget->repaint(FALSE);
			break;
		}

		case QEvent::Leave: {
			if (object != singleton->hoverWidget)
				break;
			QWidget *w = singleton->hoverWidget;
			singleton->hoverWidget = 0;
			w->repaint(FALSE);
			break;
		}

		case QEvent::MouseMove: {
			if (! object->isWidgetType() || object != singleton->hoverWidget)
				break;

			if (! object->inherits("QScrollBar") && ! object->inherits("QSlider"))
				break;

			singleton->mousePos = ((QMouseEvent *) event)->pos();
			if (! singleton->mousePressed) {
				singleton->hovering = TRUE;
				singleton->hoverWidget->repaint(FALSE);
				singleton->hovering = FALSE;
			}

			break;
		}

		default: {
			break;
		}
	}

	return QCleanlooksStyle::eventFilter(object, event);
}

void
QuarticurveStyle::drawTextRect(QPainter *p, const QRect &r,
                             const QColorGroup &cg,
                             QStyle::State flags,
                             const QBrush *fill) const
{
	QRect br = r;

	const QuarticurveColorData *cdata = lookupData(cg);

	if (flags & QStyle::State_Sunken) {
		br.addCoords(0, 0, -1, -1);
		p->save();
		p->setPen(cdata->shades[5]);
		p->setRenderHint(QPainter::Antialiasing, false);
		p->drawRect(br);

		// button bevel
		p->setPen(cg.light());
		p->drawLine(br.x() + br.width() - 2, br.y() + 2,
		            br.x() + br.width() - 2, br.y() + br.height() - 3); // right
		p->drawLine(br.x() + 2, br.y() + br.height() - 2,
		            br.x() + br.width() - 2, br.y() + br.height() - 2); // bottom

		p->setPen(cdata->shades[1]);
		p->drawLine(br.x() + 1, br.y() + 2,
		            br.x() + 1, br.y() + br.height() - 2); // left
		p->drawLine(br.x() + 1, br.y() + 1,
		            br.x() + br.width() - 2, br.y() + 1); // top

		br.addCoords(2, 2, -2, -2);
		p->restore();
	}

	// fill
	if (fill) {
		p->fillRect(br, *fill);
	}
}

void
QuarticurveStyle::drawLightBevel(QPainter *p, const QRect &r,
                               const QColorGroup &cg,
                               QStyle::State flags,
                               const QBrush *fill,
                               bool dark) const
{
	QRect br = r;
	QColor col;

	bool sunken = (flags & (QStyle::State_On | QStyle::State_Sunken));
	bool raised = (flags & (QStyle::State_Raised));

	const QuarticurveColorData *cdata = lookupData(cg);

	p->save();
	p->setPen(dark ? cdata->shades[6] : cdata->shades[5]);
	p->setRenderHint(QPainter::Antialiasing, false);
	br.addCoords(0, 0, -1, -1);
	p->drawRect(br);

	if (flags & (QStyle::State_On |
				QStyle::State_Sunken | QStyle::State_Raised)) {
		// button bevel
		p->setPen(sunken ? Qt::white : cdata->shades[2]);
		p->drawLine(r.x() + r.width() - 2, r.y() + 2,
		            r.x() + r.width() - 2, r.y() + r.height() - 3); // right
		p->drawLine(r.x() + 1, r.y() + r.height() - 2,
		            r.x() + r.width() - 2, r.y() + r.height() - 2); // bottom

		p->setPen(raised ? Qt::white : cdata->shades[2]);
		p->drawLine(r.x() + 1, r.y() + 2,
		            r.x() + 1, r.y() + r.height() - 3); // left
		p->drawLine(r.x() + 1, r.y() + 1,
		            r.x() + r.width() - 2, r.y() + 1); // top

		br.addCoords(2, 2, -1, -1);
	} else {
		br.addCoords(1, 1, -1, -1);
	}

	// fill
	if (fill) {
		p->fillRect(br, *fill);
	}
	p->restore();
}

static QPalette adjustPalette(QPalette p)
{
    /* Force the selection colors to always active. This overrides KDE 4's
       behavior of using window colors for the inactive or disabled
       selection and matches the KDE 3 and GTK+ themes. */
    QBrush highlight = p.brush(QPalette::Active, QPalette::Highlight);
    p.setBrush(QPalette::Inactive, QPalette::Highlight, highlight);
    p.setBrush(QPalette::Disabled, QPalette::Highlight, highlight);
    QBrush highlightedText = p.brush(QPalette::Active, QPalette::HighlightedText);
    p.setBrush(QPalette::Inactive, QPalette::HighlightedText, highlightedText);
    p.setBrush(QPalette::Disabled, QPalette::HighlightedText, highlightedText);
    return p;
}

void QuarticurveStyle::drawPrimitive(	PrimitiveElement pe,
					const QStyleOption *option,
					QPainter *p,
					const QWidget *widget ) const
{
    QColorGroup cg(adjustPalette(option->palette));
    QRect r(option->rect);
    QStyle::State flags(option->state);
    const QuarticurveColorData *cdata = lookupData(cg);
    switch (pe) {
		case QStyle::PE_IndicatorHeaderArrow: {
			QStyleOption changedOptions(*option);
			changedOptions.state = QStyle::State_Enabled;
			drawPrimitive((flags & QStyle::State_UpArrow) ? QStyle::PE_IndicatorArrowUp : QStyle::PE_IndicatorArrowDown, &changedOptions, p, widget);
			break;
		}

		case QStyle::PE_PanelButtonTool:
			// Disable AutoRaise for QDockWidgetTitleButton to match Qt 3 look
			// (Qt 4 sets this unconditionally)
			if (widget && widget->inherits("QDockWidgetTitleButton")) {
			  flags &= ~QStyle::State_AutoRaise;
			  if (!((flags & QStyle::State_On) || (flags & QStyle::State_Sunken)))
			    flags |= QStyle::State_Raised;
			}
			// fall through
		case QStyle::PE_PanelButtonCommand:
		case QStyle::PE_PanelButtonBevel: {
			const QBrush *fill;
			if (flags & QStyle::State_Enabled) {
			  if ((flags & (QStyle::State_On | QStyle::State_Sunken))) {
                // Pushed button is always dark. TODO: KDE buttons use a different dark
                fill = &cg.brush(QColorGroup::Dark);
              } else {
                if (flags & QStyle::State_MouseOver)
                  fill = &cg.brush(QColorGroup::Midlight);
                else
                  fill = &cg.brush(QColorGroup::Button);
              }
			} else
			  fill = &cg.brush(QColorGroup::Background);
			drawLightBevel(p, r, cg, flags, fill, true);
			break;
		}

		case QStyle::PE_IndicatorButtonDropDown: {
			QBrush thefill;
			bool sunken =
				(flags & (QStyle::State_On | QStyle::State_Sunken));

			if (flags & QStyle::State_Enabled) {
				if (sunken)
				thefill = cg.brush(QColorGroup::Midlight);
				else
				thefill = cg.brush(QColorGroup::Button);
			} else
				thefill = cg.brush(QColorGroup::Background);

			p->save();
			p->setPen(cg.dark());
			p->setRenderHint(QPainter::Antialiasing, false);
			p->drawLine(r.topLeft(),	 r.topRight());
			p->drawLine(r.topRight(),	r.bottomRight());
			p->drawLine(r.bottomRight(), r.bottomLeft());

			if (flags & (QStyle::State_On | QStyle::State_Sunken | QStyle::State_Raised)) {
				// button bevel
				if (sunken)
					p->setPen(cg.mid());
				else
					p->setPen(cg.light());

				p->drawLine(r.x(), r.y() + 2,
					r.x(), r.y() + r.height() - 3); // left
				p->drawLine(r.x(), r.y() + 1,
					r.x() + r.width() - 2, r.y() + 1); // top

				if (sunken)
					p->setPen(cg.light());
				else
					p->setPen(cg.mid());

				p->drawLine(r.x() + r.width() - 2, r.y() + 2,
					r.x() + r.width() - 2, r.y() + r.height() - 3); // right
				p->drawLine(r.x() + 1, r.y() + r.height() - 2,
					r.x() + r.width() - 2, r.y() + r.height() - 2); // bottom
			}

			p->fillRect(r.x() + 1, r.y() + 2, r.width() - 3, r.height() - 4, thefill);
			p->restore();
			break;
		}

		case QStyle::PE_FrameDefaultButton: {
			p->save();
			p->setPen(cg.shadow());
			p->setRenderHint(QPainter::Antialiasing, false);
			p->setBrush(Qt::NoBrush);
			p->drawRect(r);
			p->restore();
			break;
		}

	    case QStyle::PE_IndicatorMenuCheckMark: {
            QPoint qp = QPoint(r.center().x() - RADIO_SIZE/4, 
                               r.center().y() - RADIO_SIZE/2);
            if (flags & QStyle::State_Active) {
			    p->drawPixmap(qp, *(cdata->checkMark[0]));
            } else {
			    p->drawPixmap(qp, *(cdata->checkMark[1]));
            }
	 	    break;
	    }

		case QStyle::PE_IndicatorCheckBox: {
			int pix = 0;
			if (flags & QStyle::State_Sunken)
			  pix += 3;
			
			if (flags & QStyle::State_On)
			  pix += 1;
			else if (flags & QStyle::State_NoChange)
			  pix += 2;

			p->drawPixmap(r.topLeft(), *cdata->checkPix[pix]);
			break;
		}

#if 0
		case QStyle::PE_IndicatorMask: {
			p->fillRect(r, Qt::color1);
			break;
		}
#endif

		case QStyle::PE_IndicatorRadioButton: {
			int pix = 0;
			if (flags & QStyle::State_MouseOver) {
			  pix += 4;
			  p->fillRect(r, cg.brush(QColorGroup::Midlight));
			} else {
			  p->fillRect(r, cg.brush(QColorGroup::Background));
			}

			if (flags & QStyle::State_Sunken)
			  pix += 2;
			if (flags & QStyle::State_On)
			  pix += 1;

			p->drawPixmap(r.topLeft(), *cdata->radioPix[pix]);

			break;
		}

#if 0
		case QStyle::PE_ExclusiveIndicatorMask: {
			p->save();
			p->setPen(Qt::color1);
			p->setRenderHint(QPainter::Antialiasing, false);
			p->setBrush(Qt::color1);
			p->drawPixmap(r.topLeft(), *cdata->radioMask);
			p->restore();
			break;
		}
#endif
		

		case QStyle::PE_IndicatorToolBarHandle: {
			QString title;
			bool drawTitle = FALSE;
			if ( p && p->device()->devType() == QInternal::Widget ) {
				QWidget *w = (QWidget *) p->device();
				QWidget *p = w->parentWidget();
				if (p->inherits("Q3DockWindow") && ! p->inherits("Q3ToolBar")) {
					drawTitle = TRUE;
					title = p->caption();
				}
			}

			flags |= QStyle::State_Raised;

			if (flags & QStyle::State_Horizontal) {
				if (drawTitle) {
					QPixmap pm(r.height(), r.width());
					QPainter p2(&pm);
					p2.fillRect(0, 0, pm.width(), pm.height(),
							cg.brush(QColorGroup::Highlight));
					p2.setPen(cg.highlightedText());
					p2.drawText(0, 0, pm.width(), pm.height(), Qt::AlignCenter, title);
					p2.end();

					QMatrix m;
					m.rotate(270.0);
					pm = pm.xForm(m);
					p->drawPixmap(r.x(), r.y(), pm);
				} else {
					p->fillRect(r, cg.button());

					int yy = r.top() + 3;
					int nLines = (r.height() - 6) / 5;

					p->save();
					p->setRenderHint(QPainter::Antialiasing, false);
					for (int i = 0; i < nLines; yy += 5, i++) {
						p->setPen(cdata->shades[5]);
						p->drawLine(1, yy + 3, 4, yy);
						p->setPen(cg.light());
						p->drawLine(1, yy + 4, 4, yy + 1);
					}
					p->restore();
				}
			} else {
				p->save();
				p->setRenderHint(QPainter::Antialiasing, false);
				if (drawTitle) {
					p->fillRect(r, cg.brush(QColorGroup::Highlight));
					p->setPen(cg.highlightedText());
					p->drawText(r, Qt::AlignCenter, title);
				} else {
					p->fillRect(r, cg.button());

					int xx = r.left() + 3;
					int nLines = (r.width() - 5) / 4;

					for (int i = 0; i < nLines; xx += 4, i++) {
						p->setPen(cdata->shades[5]);
						p->drawLine(xx + 3, 1,
								xx, 1 + 3);
						p->setPen(cg.light());
						p->drawLine(xx + 3, 2,
								xx + 1, 1 + 3);
					}
				}
				p->restore();
			}
			break;
		}

		case QStyle::PE_Q3DockWindowSeparator: {
			if (r.width() > 20 || r.height() > 20) {
				p->save();
				p->setRenderHint(QPainter::Antialiasing, false);
				if (flags & QStyle::State_Horizontal) {
					p->setPen(cdata->shades[5]);
					p->drawLine(r.left() + 1, r.top() + 6, r.left() + 1, r.bottom() - 6);
					p->setPen(cdata->shades[3]);
					p->drawLine(r.left() + 2, r.top() + 6, r.left() + 2, r.bottom() - 6);
				} else {
					p->setPen(cdata->shades[5]);
					p->drawLine(r.left() + 6, r.top() + 1, r.right() - 6, r.top() + 1);
					p->setPen(cdata->shades[3]);
					p->drawLine(r.left() + 6, r.top() + 2, r.right() - 6, r.top() + 2);
				}
				p->restore();
			} else {
				QCleanlooksStyle::drawPrimitive(pe, option, p, widget);
			}
			break;
		}

		case QStyle::PE_IndicatorDockWidgetResizeHandle: {
			p->save();
			p->setRenderHint(QPainter::Antialiasing, false);
			p->fillRect(r, cg.background());
			if (flags & QStyle::State_Horizontal) {
				p->setPen(cg.highlight().light());
				p->drawLine(r.left() + 1, r.top() + 1, r.right() - 1, r.top() + 1);
				p->setPen(cg.highlight());
				p->drawLine(r.left() + 1, r.top() + 2, r.right() - 1, r.top() + 2);
				p->setPen(cg.highlight().dark());
				p->drawLine(r.left() + 1, r.top() + 3, r.right() - 1, r.top() + 3);
			} else {
				p->setPen(cg.highlight().light());
				p->drawLine(r.left() + 1, r.top() + 1, r.left() + 1, r.bottom() - 1);
				p->setPen(cg.highlight());
				p->drawLine(r.left() + 2, r.top() + 1, r.left() + 2, r.bottom() - 1);
				p->setPen(cg.highlight().dark());
				p->drawLine(r.left() + 3, r.top() + 1, r.left() + 3, r.bottom() - 1);
			}
			p->restore();
			break;
		}

		case QStyle::PE_PanelLineEdit: {
			const QStyleOptionFrame *frameOption = qstyleoption_cast<const QStyleOptionFrame *>(option);
			if (frameOption) {
				if (frameOption->lineWidth)
					flags |= QStyle::State_Sunken;
				else
					flags &= ~QStyle::State_Sunken;
			}
			drawTextRect(p, r, cg, flags, &cg.brush(QColorGroup::Base));
			break;
		}

		case QStyle::PE_FrameTabWidget: {
			QTabBar::Shape shape = QTabBar::RoundedNorth;
			const QStyleOptionTab *tabOption = qstyleoption_cast<const QStyleOptionTab *>(option);
			if (tabOption)
				shape = tabOption->shape;
			else {
				const QStyleOptionTabWidgetFrame *tabWidgetFrameOption = qstyleoption_cast<const QStyleOptionTabWidgetFrame *>(option);
				if (tabWidgetFrameOption)
					shape = tabWidgetFrameOption->shape;
				else {
					const QStyleOptionTabBarBase *tabBarBaseOption = qstyleoption_cast<const QStyleOptionTabBarBase *>(option);
					if (tabBarBaseOption)
						shape = tabBarBaseOption->shape;
				}
			}
			if (shape == QTabBar::RoundedNorth || shape == QTabBar::TriangularNorth)
				r.adjust(0, -1, 0, 0);
			else if (shape == QTabBar::RoundedSouth || shape == QTabBar::TriangularSouth)
				r.adjust(0, 0, 0, 1);
			else if (shape == QTabBar::RoundedWest || shape == QTabBar::TriangularWest)
				r.adjust(-1, 0, 0, 0);
			else if (shape == QTabBar::RoundedEast || shape == QTabBar::TriangularEast)
				r.adjust(0, 0, 1, 0);
            p->save();
            p->setPen(cdata->shades[6]);
            p->setRenderHint(QPainter::Antialiasing, false);
            p->drawLine(r.left(),r.top(),r.left(),r.bottom());
            p->drawLine(r.left(),r.bottom(),r.right(),r.bottom());
            p->drawLine(r.right(),r.bottom(),r.right(),r.top());
            p->drawLine(r.left(),r.top(),r.right(),r.top());
            p->setPen(cg.light());
            p->drawLine(r.left()+1,r.top(),r.left()+1,r.bottom()-1);
            p->drawLine(r.left()+1,r.bottom()-1,r.right()-1,r.bottom()-1);
            p->drawLine(r.right()-1,r.bottom()-1,r.right()-1,r.top()+1);
            p->drawLine(r.left(),r.top()+1,r.right()-1,r.top()+1);
            p->restore();
            break;
            
        }

		case QStyle::PE_Frame:
		case QStyle::PE_FrameWindow: {
			if ( ! ( flags & QStyle::State_Sunken ) ) {
				flags |= QStyle::State_Raised;
			}

			drawLightBevel(p, r, cg, flags);

			break;
		}

		case QStyle::PE_FrameStatusBar: {
			drawLightBevel(p, r, cg, flags | QStyle::State_Sunken);
			break;
		}

		case QStyle::PE_FrameGroupBox: {
			drawLightBevel(p, r, cg, flags | QStyle::State_Sunken | QStyle::State_Raised);
			break;
		}

		case QStyle::PE_FrameMenu: {
			drawLightBevel(p, r, cg, flags);
			break;
		}

		case QStyle::PE_FrameDockWidget:
		case QStyle::PE_PanelMenuBar: {
			p->save();
			p->setRenderHint(QPainter::Antialiasing, false);
			p->fillRect(r, cg.button());
			p->setPen(cg.mid());
			p->drawLine(r.left(), r.bottom(), r.right(), r.bottom());
			p->restore();
			break;
		}

		case QStyle::PE_FrameFocusRect: {
            p->save();
            p->setRenderHint(QPainter::Antialiasing, false);
            p->setPen(Qt::black);
            int rw = r.width(), rh = r.height(), rx = r.x(), ry = r.y();
            for (int x = 0; x < rw; x += 2) { // Top line
                p->drawPoint(rx + x, ry);
            }
            for (int y = 0; y < rh; y += 2) { // Left line
                p->drawPoint(rx, ry + y);
            }
            if (rw & 1) { // Odd width, top line one pixel short
                for (int y = 2; y < rh; y += 2) { // Right line
                    p->drawPoint(rx + rw - 1, ry + y);
                }
            } else { // Even width, top line symmetrical
                for (int y = 1; y < rh; y += 2) { // Right line
                    p->drawPoint(rx + rw - 1, ry + y);
                }
            }
            if (rh & 1) { // Odd height
                for (int x = 0; x < rw; x += 2) { // Bottom line
                    p->drawPoint(rx + x, ry + rh - 1);
                }
            } else {
                for (int x = 1; x < rw; x += 2) { // Bottom line
                    p->drawPoint(rx + x, ry + rh - 1);
                }
            }
            p->restore();
			break;
		}

		case QStyle::PE_IndicatorProgressChunk: {
			drawGradientBox(p, r, cg, cdata, false, 0.92,1.66);
			break;
		}

		case QStyle::PE_IndicatorArrowUp:
		case QStyle::PE_IndicatorArrowDown:
		case QStyle::PE_IndicatorArrowRight:
		case QStyle::PE_IndicatorArrowLeft: {
			Q3PointArray a;

			switch ( pe ) {
				case QStyle::PE_IndicatorArrowUp: {
					a.setPoints(11,   3,1,  0,-2,  -3,1,  -2,0,  -2,2,  0,-1,  2,1,  2,2,  0,0,  -1,1,  1,1);
					break;
				}

				case QStyle::PE_IndicatorArrowDown: {
					a.setPoints(11,   3,-1,  0,2,  -3,-1,  -2,0,  -2,-2,  0,1,  2,-1,  2,-2,  0,0,  -1,-1,  1,-1);
					break;
				}

				case QStyle::PE_IndicatorArrowRight: {
					a.setPoints(13,  0,-3,  -1,-2,  1,-2,  2,-1,  0,-1,  1,0,  3,0,  2,1,  0,1,  -1,2,  1,2,  0,3,  0,0);
					break;
				}

				case QStyle::PE_IndicatorArrowLeft: {
					a.setPoints(13,  0,-3,  1,-2,  -1,-2,  -2,-1,  0,-1,  -1,0,  -3,0,  -2,1,  0,1,  1,2,  -1,2,  0,3,  0,0);
					break;
				}

				default: {
					break;
				}
			}

			if (a.isNull())
				return;

			p->save();
			p->setRenderHint(QPainter::Antialiasing, false);
			a.translate( (r.x() + r.width() / 2), 
					 (r.y() + r.height() / 2));
			if ( flags & QStyle::State_Enabled ) {
				p->setPen( flags & (QStyle::State_MouseOver) ? cg.highlightedText() : cg.buttonText() );
			} else {
				p->setPen(cdata->shades[7]);
			}
			p->drawPolyline(a);
			p->restore();

			break;
		}

		case QStyle::PE_IndicatorSpinUp:
		case QStyle::PE_IndicatorSpinDown: {
			p->save();
			p->setPen(cdata->shades[7]);
			p->setRenderHint(QPainter::Antialiasing, false);
            p->setBrush(Qt::NoBrush);
			Q3PointArray a;
			if (pe==QStyle::PE_IndicatorSpinUp) {
				a.setPoints(8,  -3,0,  -1,-2,  1,0,  1,1,  -1,-1,  -3,1,  -3,0,  1,0);
			} else {
				a.setPoints(8,  -3,-1,  -1,1,  1,-1,  1,-2,  -1,0,  -3,-2,  -3,-1,  1,-1);
			}
			a.translate(r.x() + r.width() / 2, r.y() + r.height() / 2 + 1);
			p->drawPolyline(a);
			p->restore();
			break;
		}

		default: {
			QCleanlooksStyle::drawPrimitive(pe, option, p, widget);
			break;
		}
    }
}

void QuarticurveStyle::drawControl(ControlElement control,
				  const QStyleOption *option,
				  QPainter *p,
				  const QWidget *widget) const
{
    QColorGroup cg(adjustPalette(option->palette));
    QRect r(option->rect);
    QStyle::State flags(option->state);
	const QuarticurveColorData *cdata = lookupData(cg);

	if (widget && widget == singleton->hoverWidget) {
		flags |= QStyle::State_MouseOver;
	}

	switch (control) {

		case QStyle::CE_HeaderSection: {
			const QBrush *fill;
			if (! (flags & QStyle::State_Sunken) && (flags & QStyle::State_On))
				fill = &cg.brush(QColorGroup::Midlight);
			else
				fill = &cg.brush(QColorGroup::Button);
			flags = ((flags | QStyle::State_Sunken) ^ QStyle::State_Sunken) | QStyle::State_Raised;
			drawLightBevel(p, r, cg, flags, fill);
			/* p->setPen( cg.buttonText() ); */
			break;
		}

		case QStyle::CE_Splitter: {
			p->save();
			p->setRenderHint(QPainter::Antialiasing, false);
			if (flags & QStyle::State_Horizontal) {
			  int y_mid = r.height()/2;
			  for (int i=0; i< 21; i=i+5) {
				p->setPen(cdata->shades[5]);
				p->drawLine(r.x()+1, y_mid-10+i, r.right()-1, y_mid-10+i-3);
				p->setPen(cg.light());
				p->drawLine(r.x()+1, y_mid-10+i+1, r.right()-1, y_mid-10+i-2);
			  }
			} else {
			  int x_mid = r.width()/2;
			  for (int i=0; i< 21; i=i+5) {
				p->setPen(cdata->shades[5]);
				p->drawLine(x_mid-10+i+3, r.y()+1, x_mid-10+i, r.bottom()-1);
				p->setPen(cg.light());
				p->drawLine(x_mid-10+i+4, r.y()+1, x_mid-10+i+1, r.bottom()-1);
			  }
			}
				
			if (flags & QStyle::State_Horizontal) {
				flags &= ~QStyle::State_Horizontal;
			} else {
				flags |= QStyle::State_Horizontal;
			}
			p->restore();
			break;
		}

		case QStyle::CE_ScrollBarAddLine:
		case QStyle::CE_ScrollBarSubLine: {
            // Highlight on mouse over
			if (widget) {
				QPoint mp = widget->mapFromGlobal(widget->cursor().pos());
				if (!r.contains(mp))
					flags &= ~QStyle::State_MouseOver;
			}
			drawLightBevel(p, r, cg, ((flags | QStyle::State_Sunken) ^ QStyle::State_Sunken) |
					   ((flags & QStyle::State_Enabled) ? QStyle::State_Raised : QStyle::State_Default),
					   &cg.brush((flags & QStyle::State_MouseOver) ? QColorGroup::Midlight : QColorGroup::Button), true);

			PrimitiveElement pe;
			if ((control == QStyle::CE_ScrollBarAddLine) && (flags & QStyle::State_Horizontal)) {
				pe = QStyle::PE_IndicatorArrowRight;
			} else if (control == QStyle::CE_ScrollBarAddLine) {
				pe = QStyle::PE_IndicatorArrowDown;
			} else if (flags & QStyle::State_Horizontal) {
				pe = QStyle::PE_IndicatorArrowLeft;
			} else {
				pe = QStyle::PE_IndicatorArrowUp;
			}
            // But the arrow is always black
			QStyleOption changedOptions(*option);
			changedOptions.state &= ~QStyle::State_MouseOver;
			drawPrimitive(pe, &changedOptions, p, widget);
			break;
		}

		case QStyle::CE_ScrollBarSubPage:
		case QStyle::CE_ScrollBarAddPage: {
			p->save();
			p->setRenderHint(QPainter::Antialiasing, false);
			p->fillRect(r, cdata->shades[3]);
			p->setPen(cdata->shades[5]);
			if (flags & QStyle::State_Horizontal) {
			  p->drawLine(r.left(), r.top(), r.right(), r.top());
			  p->drawLine(r.left(), r.bottom(), r.right(), r.bottom());
			} else {
			  p->drawLine(r.left(), r.top(), r.left(), r.bottom());
			  p->drawLine(r.right(), r.top(), r.right(), r.bottom());
			}
			p->restore();
			break;
		}

		case QStyle::CE_ScrollBarSlider: {
			int x1, y1;
			p->save();
			p->setPen(cg.dark());
			p->setRenderHint(QPainter::Antialiasing, false);
            // Not the best place to put this, but it seems to be the only place where it works correctly

            // Expand the slider one pixel in each direction, 
            // to avoid double lines in the extreme positions.
            QRect r2 = r;
            if (flags & QStyle::State_Horizontal) {
                r2.setWidth(r.width() + 1);
                r2.setX(r.x() - 1);
            } else {
                r2.setHeight(r.height() + 1);
                r2.setY(r.y() - 1);
            }
            // Highlight on mouse over
			if (widget) {
				QPoint mp = widget->mapFromGlobal(widget->cursor().pos());
				if (!r.contains(mp))
					flags &= ~QStyle::State_MouseOver;
			}
			drawLightBevel(p, r2, cg, 
                           ((flags | QStyle::State_Sunken) ^ QStyle::State_Sunken) | ((flags & QStyle::State_Enabled) ? QStyle::State_Raised : QStyle::State_Default),
					       &cg.brush((flags & QStyle::State_MouseOver || flags & QStyle::State_Sunken) ? QColorGroup::Midlight : QColorGroup::Button), true);

			if (flags & QStyle::State_Horizontal && r.width() < 31) {
				p->restore();
				break;
			}
			if (!(flags & QStyle::State_Horizontal) && r.height() < 31) {
				p->restore();
				break;
			}
		
            // Scrollbar diagonal handle
			p->setPen(cdata->shades[5]);
			if (flags & QStyle::State_Horizontal) {
				x1 = (r.left() + r.right()) / 2 - 8;
				y1 = ((r.top() + r.bottom()) - 6) / 2;
				p->drawLine(x1 + 5, y1, x1, y1 + 5);
				p->drawLine(x1 + 5 + 5, y1,	x1 + 5, y1 + 5);
				p->drawLine(x1 + 5 + 5*2, y1, x1 + 5*2, y1 + 5);
			} else {
				x1 = ((r.left() + r.right()) - 6) / 2;
				y1 = (r.top() + r.bottom()) / 2 - 8;
				p->drawLine(x1 + 5, y1,	x1, y1 + 5);
				p->drawLine(x1 + 5, y1 + 5,	x1, y1 + 5 + 5);
				p->drawLine(x1 + 5, y1 + 5*2, x1, y1 + 5 + 5*2);
			}

			p->setPen(Qt::white);
			if (flags & QStyle::State_Horizontal) {
				x1 = (r.left() + r.right()) / 2 - 8;
				y1 = ((r.top() + r.bottom()) - 6) / 2;
				p->drawLine(x1 + 5, y1+1, x1 + 1, y1 + 5);
				p->drawLine(x1 + 5 + 5, y1 + 1,	x1 + 1 + 5, y1 + 5);
				p->drawLine(x1 + 5 + 5*2, y1 + 1, x1 + 1 + 5*2, y1 + 5);
			} else {
				x1 = ((r.left() + r.right()) - 6) / 2;
				y1 = (r.top() + r.bottom()) / 2 - 8;
				p->drawLine(x1 + 5, y1 + 1,	x1 + 1, y1 + 5);
				p->drawLine(x1 + 5, y1 + 1 + 5,	x1 + 1, y1 + 5 + 5);
				p->drawLine(x1 + 5, y1 + 1 + 5*2, x1 + 1, y1 + 5 + 5*2);
			}
			p->restore();

			break;
		}

        // Mozer
        case CE_PushButtonLabel: { // Case based on qcommonstyle.cpp
            const QStyleOptionButton *buttonOption = qstyleoption_cast<const QStyleOptionButton *>(option);
            if (!buttonOption) { // oops, panic
                QCleanlooksStyle::drawControl(control, option, p, widget);
                break;
            }

            QRect ir = r;
            if (flags & QStyle::State_On)
                flags |= QStyle::State_Sunken;
            if (flags & QStyle::State_Sunken)
                ir.moveBy(pixelMetric(PM_ButtonShiftHorizontal, option, widget), pixelMetric(PM_ButtonShiftVertical, option, widget));

            if (buttonOption->features & QStyleOptionButton::HasMenu) {
                int mbi = pixelMetric(PM_MenuButtonIndicator, option, widget);
#if 0 /* Qt 4 does this automatically */
                QRect ar(ir.right() - mbi, ir.y() + 2, mbi - 4, ir.height() - 4);
                QStyleOption changedOptions(*option);
                changedOptions.rect = ar;
                drawPrimitive(QStyle::PE_IndicatorArrowDown, &changedOptions, p, widget);
#endif
                ir.setWidth(ir.width() - mbi);
            }

            int tf=Qt::AlignVCenter | Qt::TextShowMnemonic;
            if (!styleHint(SH_UnderlineAccelerator, 0, widget, 0))
                tf |= Qt::TextHideMnemonic;

            if ( ! buttonOption->icon.isNull() ) {
                QIcon::Mode mode =
                    (flags & QStyle::State_Enabled) ? QIcon::Normal : QIcon::Disabled;
                if ( mode == QIcon::Normal && (flags & QStyle::State_HasFocus) )
                    mode = QIcon::Active;

                QIcon::State state = QIcon::Off;
                if ( flags & QStyle::State_On )
                    state = QIcon::On;

                QPixmap pixmap = buttonOption->icon.pixmap( buttonOption->iconSize, mode, state );
                int pixw = pixmap.width();
                int pixh = pixmap.height();

                // content rectangle is pixmap + spacer + text bounding rectangle

                int csp = 5; // content spacer (between pixmap and text)
                QFontMetrics fm = p->fontMetrics();
                int contw;
                if (!buttonOption->text.isEmpty())
                    contw = pixw + fm.width(buttonOption->text) + csp; // width of pixmap + text
                else
                    contw = pixw;
                int contx = ir.x() + ir.width()/2 - contw/2; // upper left corner of content rectangle

                //Center the icon if there is no text
                if ( buttonOption->text.isEmpty() )
                    p->drawPixmap( ir.x() + ir.width() / 2 - pixw / 2, ir.y() + ir.height() / 2 - pixh / 2, pixmap );
                else 
                    p->drawPixmap( contx, ir.y() + ir.height() / 2 - pixh / 2, pixmap );
                ir.setLeft(contx);
                ir.setWidth(contw);
                
                // for debugging
                //p->setPen(Qt::black); p->drawRect(ir);
                
                // right-align text in the content rectangle, if there is any
                if (!buttonOption->text.isEmpty())
                    tf |= Qt::AlignRight;
                //else if (button->pixmap())
                //    tf |= Qt::AlignHCenter;
            } else {
                tf |= Qt::AlignHCenter;
            }
            drawItem(p, ir, tf, cg,
                    flags & QStyle::State_Enabled, 0, buttonOption->text,
                    QPalette::ButtonText );

#if 0 // now drawn automatically
            if (flags & QStyle::State_HasFocus) {
                QStyleOption changedOptions(*option);
                changedOptions.rect = subElementRect(QStyle::SE_PushButtonFocusRect, option, widget);
                drawPrimitive(QStyle::PE_FrameFocusRect, &changedOptions, p, widget);
            }
#endif
            break;
        }

        
		case CE_TabBarTabShape: {
			QTabBar::Shape shape = QTabBar::RoundedNorth;
			const QStyleOptionTab *tabOption = qstyleoption_cast<const QStyleOptionTab *>(option);
			if (tabOption)
				shape = tabOption->shape;
			else {
				const QStyleOptionTabWidgetFrame *tabWidgetFrameOption = qstyleoption_cast<const QStyleOptionTabWidgetFrame *>(option);
				if (tabWidgetFrameOption)
					shape = tabWidgetFrameOption->shape;
				else {
					const QStyleOptionTabBarBase *tabBarBaseOption = qstyleoption_cast<const QStyleOptionTabBarBase *>(option);
					if (tabBarBaseOption)
						shape = tabBarBaseOption->shape;
				}
			}
			bool below = false;
			QRect tr(r);
			QRect fr(r);

			if (shape == QTabBar:: RoundedEast || shape == QTabBar:: TriangularEast
			    || shape == QTabBar:: RoundedWest || shape == QTabBar:: TriangularWest) {
				tr.addCoords(0, 0,  -1, 0);
				fr.addCoords(2, 2, 0, -2);

				if ( shape == QTabBar:: RoundedEast || shape == QTabBar:: TriangularEast) {
					tr = r; tr.addCoords(1, 0, 0, 0);
					fr = r; fr.addCoords(0, 2,-2, -2);
					below = true;
				}    

				if (!widget || r.bottom() != widget->rect().bottom()) {
					tr.addCoords(0, 0, 0, 1);
					fr.addCoords(0, 0, 0, 1);
				}

				p->save();
				p->setRenderHint(QPainter::Antialiasing, false);
				if (! (flags & QStyle::State_Selected)) {
					if (below) {
						tr.addCoords(0, 0, -1, 0);
						fr.addCoords(2, 0, -1, 0);
					} else {
						tr.addCoords(1, 0, 0, 0);
						fr.addCoords(1, 0, -2, 0);
					}

					p->setPen(cdata->shades[6]);
					tr.addCoords(0, 0, -1, -1);
					p->drawRect(tr);
					tr.addCoords(0, 0, 1, 1);

					if (tr.top() == 0)
						if (below)
							p->drawPoint(tr.left() - 1, tr.top());
						else
							p->drawPoint(tr.right() + 1, tr.top());

					p->setPen(cg.light());
					if (!below) {
						p->drawLine(tr.left() + 1, tr.top() + 1, tr.left() + 1, tr.bottom() - 1);
						if (tr.top() == 0)
							p->drawLine(tr.right() + 1, tr.top() + 1,	tr.right() + 1, tr.bottom());
						else
							p->drawLine(tr.right() + 1, tr.top(),	tr.right() + 1, tr.bottom());
					}

					p->setPen(cdata->shades[2]);
					if (below) {
						p->drawLine(tr.right() - 1, tr.top() + 1,	tr.right() - 1, tr.bottom() - 1);
						if (tr.top() == 0)
							p->drawLine(tr.left() - 1, tr.top() + 1, tr.left() - 1, tr.bottom());
						else
							p->drawLine(tr.left() - 1, tr.top(), tr.left() - 1, tr.bottom());
						p->drawLine(tr.left() + 1, tr.top() + 1, tr.right() - 2, tr.top() + 1);
						p->drawLine(tr.left() + 1, tr.bottom() - 1, tr.right() - 2, tr.bottom() - 1);
					} else {
						p->drawLine(tr.left() + 2, tr.top() + 1, tr.right() - 1, tr.top() + 1);
						p->drawLine(tr.left() + 2, tr.bottom() - 1, tr.right() - 1, tr.bottom() - 1);
					}
				} else {
					p->setPen(cdata->shades[6]);
					if (below) {
						p->drawLine(tr.right() - 1, tr.top(),	tr.left() - 1, tr.top());
						p->drawLine(tr.right(), tr.top(),	tr.right(), tr.bottom());
						p->drawLine(tr.left(), tr.bottom(), tr.right() - 1, tr.bottom());
					} else {
						p->drawLine(tr.right() + 1, tr.top(), tr.left() + 1, tr.top());
						p->drawLine(tr.left(), tr.top(), tr.left(), tr.bottom());
						p->drawLine(tr.left() + 1, tr.bottom(), tr.right(), tr.bottom());
					}

					p->setPen(cg.light());
					if (below) {
						p->drawLine(tr.right() - 1, tr.top() + 1,	tr.left() - 2, tr.top() + 1);
						p->drawPoint(tr.left() - 1, tr.bottom());
					} else {
						p->drawLine(tr.right() + 2, tr.top() + 1,	tr.left() + 2, tr.top() + 1);
						if (tr.top() != 0)
							p->drawPoint(tr.right() + 1, tr.top());

						p->drawLine(tr.left() + 1, tr.top() + 1, tr.left() + 1, tr.bottom() - 1);
						p->drawPoint(tr.right() + 1, tr.bottom());
					}

					p->setPen(cdata->shades[2]);
					if (below) {
						if (tr.top() != 0)
							p->drawPoint(tr.left() - 1, tr.top());
						p->drawLine(tr.right() - 1, tr.top() + 2,	tr.right() - 1, tr.bottom() - 1);
						p->drawLine(tr.left() - 1, tr.bottom() - 1, tr.right() - 2, tr.bottom() - 1);
					} else {
						p->drawLine(tr.left() + 2, tr.bottom() - 1, tr.right() + 1, tr.bottom() - 1);
					}
				}

				p->fillRect(fr, ((flags & QStyle::State_Selected) ?	cg.background() : cdata->shades[2]));
				p->restore();
				break;
			}

			tr.addCoords(0, 0,  0, -1);
			fr.addCoords(2, 2, -2, 0);

			if ( shape == QTabBar:: RoundedSouth || shape == QTabBar:: TriangularSouth) {
				tr = r; tr.addCoords(0, 1, 0, 0);
				fr = r; fr.addCoords(2, 0,-2, -2);
				below = true;
			}    

			if (!widget || r.right() != widget->rect().right()) {
				tr.addCoords(0, 0, 1, 0);
				fr.addCoords(0, 0, 1, 0);
			}

			p->save();
			p->setRenderHint(QPainter::Antialiasing, false);
			if (! (flags & QStyle::State_Selected)) {
				if (below) {
					tr.addCoords(0, 0, 0, -1);
					fr.addCoords(0, 2, 0, -1);
				} else {
					tr.addCoords(0, 1, 0, 0);
					fr.addCoords(0, 1, 0, -2);
				}

				p->setPen(cdata->shades[6]);
				tr.addCoords(0, 0, -1, -1);
				p->drawRect(tr);
				tr.addCoords(0, 0, 1, 1);

				if (tr.left() == 0)
					if (below)
						p->drawPoint(tr.left(), tr.top() - 1);
					else
						p->drawPoint(tr.left(), tr.bottom() + 1);

				p->setPen(cg.light());
				if (!below) {
					p->drawLine(tr.left() + 1, tr.top() + 1, tr.right() - 1, tr.top() + 1);
					if (tr.left() == 0)
						p->drawLine(tr.left() + 1, tr.bottom() + 1,	tr.right(), tr.bottom() + 1);
					else
						p->drawLine(tr.left(), tr.bottom() + 1,	tr.right(), tr.bottom() + 1);
				}

				p->setPen(cdata->shades[2]);
				if (below) {
					p->drawLine(tr.left() + 1, tr.bottom() - 1,	tr.right() - 1, tr.bottom() - 1);
					if (tr.left() == 0)
						p->drawLine(tr.left() + 1, tr.top() - 1, tr.right(), tr.top() - 1);
					else
						p->drawLine(tr.left(), tr.top() - 1, tr.right(), tr.top() - 1);
					p->drawLine(tr.left() + 1, tr.top() + 1, tr.left() + 1, tr.bottom() - 2);
					p->drawLine(tr.right() - 1, tr.top() + 1, tr.right() - 1, tr.bottom() - 2);
				} else {
					p->drawLine(tr.left() + 1, tr.top() + 2, tr.left() + 1, tr.bottom() - 1);
					p->drawLine(tr.right() - 1, tr.top() + 2, tr.right() - 1, tr.bottom() - 1);
				}
			} else {
				p->setPen(cdata->shades[6]);
				if (below) {
					p->drawLine(tr.left(), tr.bottom() - 1,	tr.left(), tr.top() - 1);
					p->drawLine(tr.left(), tr.bottom(),	tr.right(), tr.bottom());
					p->drawLine(tr.right(), tr.top(), tr.right(), tr.bottom() - 1);
				} else {
					p->drawLine(tr.left(), tr.bottom() + 1, tr.left(), tr.top() + 1);
					p->drawLine(tr.left(), tr.top(), tr.right(), tr.top());
					p->drawLine(tr.right(), tr.top() + 1, tr.right(), tr.bottom());
				}

				p->setPen(cg.light());
				if (below) {
					p->drawLine(tr.left() + 1, tr.bottom() - 1,	tr.left() + 1, tr.top() - 2);
					p->drawPoint(tr.right(), tr.top() - 1);
				} else {
					p->drawLine(tr.left() + 1, tr.bottom() + 2,	tr.left() + 1, tr.top() + 2);
					if (tr.left() != 0)
						p->drawPoint(tr.left(), tr.bottom() + 1);

					p->drawLine(tr.left() + 1, tr.top() + 1, tr.right() - 1, tr.top() + 1);
					p->drawPoint(tr.right(), tr.bottom() + 1);
				}

				p->setPen(cdata->shades[2]);
				if (below) {
					if (tr.left() != 0)
						p->drawPoint(tr.left(), tr.top() - 1);
					p->drawLine(tr.left() + 2, tr.bottom() - 1,	tr.right() - 1, tr.bottom() - 1);
					p->drawLine(tr.right() - 1, tr.top() - 1, tr.right() - 1, tr.bottom() - 2);
				} else {
					p->drawLine(tr.right() - 1, tr.top() + 2, tr.right() - 1, tr.bottom() + 1);
				}
			}

			p->fillRect(fr, ((flags & QStyle::State_Selected) ?	cg.background() : cdata->shades[2]));
			p->restore();
			break;
		}

#if 0 /* let the base class draw this one so tabs with icons work */
        case CE_TabBarTabLabel: {
           const QStyleOptionTab * t = qstyleoption_cast<const QStyleOptionTab *>(option);
           if ( !t )
               break;

           QRect tr = r;
           /*
           if ( t->identifier() == tb->currentTab() )
               tr.setBottom( tr.bottom() - pixelMetric( QStyle::PM_DefaultFrameWidth, option, tb ) );
           */
           int alignment = Qt::AlignLeft | Qt::TextShowMnemonic;
           if (!styleHint(SH_UnderlineAccelerator, 0, widget, 0))
               alignment |= Qt::TextHideMnemonic;
            
           // Move the text bounding rectangle to place the text correctly
           tr.moveBy(5, 5);
           drawItem( p, tr, alignment, cg, flags & QStyle::State_Enabled, 0, t->text );
           // Now move it back for the focus rectangle (yes, it's a a hack)
           tr.adjust(-2, -2, -8, -8);
           if ( (flags & QStyle::State_HasFocus) && !t->text.isEmpty() ) {
               QStyleOption changedOptions(*option);
               changedOptions.rect = tr;
               drawPrimitive(QStyle::PE_FrameFocusRect, &changedOptions, p, widget);
           }
           break;
       }
#endif


		case CE_MenuItem:
		{
			const QStyleOptionMenuItem *mi = qstyleoption_cast<const QStyleOptionMenuItem *>(option);

			if ( mi && mi->menuItemType == QStyleOptionMenuItem::Separator ) {
				// draw separator
				p->save();
				p->setRenderHint(QPainter::Antialiasing, false);
				p->fillRect(r, cg.brush(QColorGroup::Button));
                p->setPen(cdata->shades[2]);
				p->drawLine(r.left() + 6, r.top() + 4, r.right() - 6, r.top() + 4);
				p->setPen(cg.light());
				p->drawLine(r.left() + 6,  r.top() + 5, r.right() - 6, r.top() + 5);
				p->restore();
				break;
			}

			if ((flags & QStyle::State_Selected) && (flags & QStyle::State_Enabled)) {
				drawGradientBox(p, r, cg, cdata, false, 0.9, 1.2);
			} else {
				p->fillRect(r, cg.brush(QColorGroup::Button));
			}

			if ( !mi )
				break;
			int tab = mi->tabWidth;
			int maxpmw = mi->maxIconWidth;

			maxpmw = QMAX(maxpmw, 22);

			QRect cr, ir, tr, sr;
			// check column
			cr.setRect(r.left(), r.top(), maxpmw, r.height());
			// submenu indicator column
			sr.setCoords(r.right() - 12, r.top(), r.right(), r.bottom());
			// tab/accelerator column
			tr.setCoords(r.right() - tab - 4, r.top(), r.right() - 4, r.bottom());
			// item column
			ir.setCoords(cr.right() + 4, r.top(), tr.right() - 4, r.bottom());

			Qt::LayoutDirection reverse = option->direction;
			if ( reverse ) {
				cr = visualRect( reverse, r, cr );
				sr = visualRect( reverse, r, sr );
				tr = visualRect( reverse, r, tr );
				ir = visualRect( reverse, r, ir );
			}

			if (!mi->icon.isNull()) {
				QIcon::Mode mode =
					(flags & QStyle::State_Enabled) ? QIcon::Normal : QIcon::Disabled;
				if ((flags & QStyle::State_Selected) && (flags & QStyle::State_Enabled))
					mode = QIcon::Active;
				QPixmap pixmap;
				if (mi->checkType != QStyleOptionMenuItem::NotCheckable && mi->checked)
					pixmap=mi->icon.pixmap( 16, mode, QIcon::On );
				else
					pixmap=mi->icon.pixmap( 16, mode );
				QRect pmr(QPoint(0, 0), pixmap.size());
				pmr.moveCenter(cr.center());
				p->save();
				p->setPen(cg.text());
				p->setRenderHint(QPainter::Antialiasing, false);
				p->drawPixmap(pmr.topLeft(), pixmap);
				p->restore();
			} else if (mi->checkType != QStyleOptionMenuItem::NotCheckable && mi->checked) {
				QStyleOption changedOptions(*option);
				changedOptions.rect = cr;
				changedOptions.state = (flags & (QStyle::State_Enabled|QStyle::State_Selected)) | QStyle::State_On;
				drawPrimitive(QStyle::PE_IndicatorMenuCheckMark, &changedOptions, p, widget);
            }
			QColor textcolor;
			QColor embosscolor;
			if (flags & QStyle::State_Selected) {
				if (! (flags & QStyle::State_Enabled)) {
					textcolor = cg.text();
					embosscolor = cg.light();
				} else {
					textcolor = cg.highlightedText();
					embosscolor = cg.midlight().light();
				}
			} else if (! (flags & QStyle::State_Enabled)) {
				textcolor = cg.text();
				embosscolor = cg.light();
			} else
				textcolor = embosscolor = cg.buttonText();
			p->save();
			p->setPen(textcolor);
			p->setRenderHint(QPainter::Antialiasing, false);

#if 0 /* no such option anymore? (Kevin Kofler) */
			if (mi->custom()) {
				if (! (flags & QStyle::State_Enabled)) {
					p->setPen(cg.light());
					mi->custom()->paint(p, cg, (flags & QStyle::State_Enabled) ? (flags & QStyle::State_Selected) : 0,
							flags & QStyle::State_Enabled,
							ir.x() + 1, ir.y() + 1,
							ir.width() - 1, ir.height() - 1);
					p->setPen(textcolor);
				}
				mi->custom()->paint(p, cg, (flags & QStyle::State_Enabled) ? (flags & QStyle::State_Selected) : 0,
						flags & QStyle::State_Enabled,
						ir.x(), ir.y(),
						ir.width(), ir.height());
			}
#endif

			QString text = mi->text;
			if (! text.isNull()) {
				int t = text.find('\t');

				// draw accelerator/tab-text
				if (t >= 0) {
					int alignFlag = Qt::AlignVCenter | Qt::TextShowMnemonic | Qt::TextDontClip | Qt::TextSingleLine;
					alignFlag |= ( reverse ? Qt::AlignLeft : Qt::AlignRight );
					if (! (flags & QStyle::State_Enabled)) {
						p->setPen(embosscolor);
						tr.moveBy(1, 1);
						p->drawText(tr, alignFlag, text.mid(t + 1));
						tr.moveBy(-1, -1);
						p->setPen(textcolor);
					}

					p->drawText(tr, alignFlag, text.mid(t + 1));
				}

				int alignFlag = Qt::AlignVCenter | Qt::TextShowMnemonic | Qt::TextDontClip | Qt::TextSingleLine;
				alignFlag |= ( reverse ? Qt::AlignRight : Qt::AlignLeft );

				if (! (flags & QStyle::State_Enabled)) {
					p->setPen(embosscolor);
					ir.moveBy(1, 1);
					p->drawText(ir, alignFlag, text, t);
					ir.moveBy(-1, -1);
					p->setPen(textcolor);
				}

				p->drawText(ir, alignFlag, text, t);
#if 0 /* apparently we have only icon instead of iconSet and pixmap now (Kevin Kofler) */
			} else if (mi->pixmap()) {
				QPixmap pixmap = *mi->pixmap();
				if (pixmap.depth() == 1)
					p->setBackgroundMode(Qt::OpaqueMode);
				p->drawPixmap(ir.x(), ir.y() + (ir.height() - pixmap.height()) / 2, pixmap);
				if (pixmap.depth() == 1)
					p->setBackgroundMode(Qt::TransparentMode);
#endif
			}
			p->restore();

			if (mi->menuItemType == QStyleOptionMenuItem::SubMenu) {
				QStyleOption changedOptions(*option);
				changedOptions.rect = sr;
				drawPrimitive((reverse ? QStyle::PE_IndicatorArrowLeft : QStyle::PE_IndicatorArrowRight), &changedOptions, p, widget);
			}
			break;
		}

		case CE_MenuBarEmptyArea: {
			p->fillRect(r, cg.brush(QColorGroup::Button));
			break;
		}

		case CE_Q3DockWindowEmptyArea: {
			p->fillRect(r, cg.brush(QColorGroup::Button));
			break;
		}

		case CE_MenuBarItem: {
		if ((flags & QStyle::State_Enabled) && (flags & QStyle::State_Selected) &&
				(flags & QStyle::State_Sunken))
			drawGradientBox(p, r, cg, cdata, false, 0.9, 1.2);
		else
			p->fillRect(r, cg.brush(QColorGroup::Button));

			const QStyleOptionMenuItem *mi = qstyleoption_cast<const QStyleOptionMenuItem *>(option);
			if ( !mi )
				break;

			if (flags & QStyle::State_Selected && (flags & QStyle::State_Sunken))
				drawItem(p, r, Qt::AlignCenter | Qt::TextShowMnemonic | Qt::TextDontClip | Qt::TextSingleLine, cg, flags & QStyle::State_Enabled, 0 /*mi->pixmap*/, mi->text, QPalette::HighlightedText);
			else
				drawItem(p, r, Qt::AlignCenter | Qt::TextShowMnemonic | Qt::TextDontClip | Qt::TextSingleLine, cg, flags & QStyle::State_Enabled, 0 /*mi->pixmap*/, mi->text, QPalette::ButtonText);
			break;
		}

		case CE_ProgressBarGroove: {
			p->save();
			p->setBrush(cdata->shades[3]);
			p->setPen(cdata->shades[5]);
			p->setRenderHint(QPainter::Antialiasing, false);
			QRect r2(r);
			r2.adjust(0, 0, -1, -1);
			p->drawRect(r2);
			p->restore();
			break;
		}

		case CE_ProgressBarContents: {
			const QStyleOptionProgressBar *progressbar = qstyleoption_cast<const QStyleOptionProgressBar *>(option);
			bool reverse = QApplication::reverseLayout();
			int totalSteps = progressbar->maximum - progressbar->minimum;

			QRect pr;

			if ( !totalSteps ) {
				int w, remains;

				// draw busy indicator

				w = MIN (25, r.width()/2);
				w = MAX (w, 1);

				remains = r.width() - w;
				remains = MAX (remains, 1);
				int x = progressbar->progress % (remains * 2);
				if (x > remains)
					x = 2 * remains - x;

				x = reverse ? r.right() - x - w : x + r.left();
				pr.setRect (x, r.top(), w, r.height());
			} else {
				int pos = progressbar->progress - progressbar->minimum;
				if (pos < 0) pos = 0;
				int total = totalSteps ? totalSteps : 1;
				int w = (int)(((double)pos*r.width())/total);

				if (reverse)
					pr.setRect (r.right() - w, r.top(), w, r.height());
				else
					pr.setRect (r.left(), r.top(), w, r.height());
			   }
			drawGradientBox(p, pr, cg, cdata, false, 0.92,1.66);

			break;
		}

#if 0 /* not useful in Qt 4 */
		case CE_CheckBox: {
		   drawPrimitive(QStyle::PE_IndicatorCheckBox, option, p, widget);
		   break;
		}
#endif

#if 0 /* let the base class handle this - there isn't really a reason to draw this specially and it breaks support for having both pixmap and text */
		case CE_CheckBoxLabel: {
			const QCheckBox *checkbox = (const QCheckBox *) widget;

#if 0 /* now drawn automatically */
			if (flags & QStyle::State_MouseOver) {
				QRegion r(checkbox->rect());
				r -= visualRect(widget->layoutDirection(), widget->rect(), subElementRect(QStyle::SE_CheckBoxIndicator, option, widget));
				p->setClipRegion(r);
				p->fillRect(checkbox->rect(), cg.brush(QColorGroup::Midlight));
				p->setClipping(FALSE);
			}
#endif
			int alignment = QApplication::reverseLayout() ? Qt::AlignRight : Qt::AlignLeft;
			drawItem(p, r, alignment | Qt::AlignVCenter | Qt::TextShowMnemonic, cg,
					flags & QStyle::State_Enabled, checkbox->pixmap(), checkbox->text());

#if 0 /* now drawn automatically */
			if (checkbox->hasFocus()) {
				QStyleOption changedOptions(*option);
				changedOptions.rect = visualRect(widget->layoutDirection(), widget->rect(), subElementRect(QStyle::SE_CheckBoxFocusRect, option, widget));
				drawPrimitive(QStyle::PE_FrameFocusRect, &changedOptions, p, widget);
			}
#endif

			break;
		}

		case CE_RadioButtonLabel: {
			const QRadioButton *radiobutton = (const QRadioButton *) widget;

#if 0 /* now drawn automatically */
			if (flags & QStyle::State_MouseOver) {
				QRegion r(radiobutton->rect());
				r -= visualRect(widget->layoutDirection(), widget->rect(), subElementRect(QStyle::SE_RadioButtonIndicator, option, widget));
				p->setClipRegion(r);
				p->fillRect(radiobutton->rect(), cg.brush(QColorGroup::Midlight));
				p->setClipping(FALSE);
			}
#endif

			int alignment = QApplication::reverseLayout() ? Qt::AlignRight : Qt::AlignLeft;
			drawItem(p, r, alignment | Qt::AlignVCenter | Qt::TextShowMnemonic, cg,
					flags & QStyle::State_Enabled, radiobutton->pixmap(), radiobutton->text());

#if 0 /* now drawn automatically */
			if (radiobutton->hasFocus()) {
				QStyleOption changedOptions(*option);
				changedOptions.rect = visualRect(widget->layoutDirection(), widget->rect(), subElementRect(QStyle::SE_RadioButtonFocusRect, option, widget));
				drawPrimitive(QStyle::PE_FrameFocusRect, &changedOptions, p, widget);
			}
#endif
			break;
		}
#endif

		case CE_DockWidgetTitle:
		   // This code adapted from QCommonStyle - Copyright (C) 1992-2007 Trolltech ASA. All rights reserved. - GPLv2
		   if (const QStyleOptionDockWidget *dwOpt = qstyleoption_cast<const QStyleOptionDockWidget *>(option)) {
		     const QStyleOptionDockWidgetV2 *v2
		       = qstyleoption_cast<const QStyleOptionDockWidgetV2*>(option);
		     bool verticalTitleBar = v2 == 0 ? false : v2->verticalTitleBar;

		     if (verticalTitleBar)
		       r.adjust(0, 0, -1, 0);
		     else
		       r.adjust(0, 0, 0, -1);
		     bool floating = false;
		     if (widget && widget->isWindow()) {
		       floating = true;
		       if (dwOpt->state & State_Active)
		         p->fillRect(r, cg.brush(QColorGroup::Highlight));
		       else
		         p->fillRect(r, cdata->shades[5]);
		     } else {
		       // Make space for buttons
		       int offset = (dwOpt->floatable && dwOpt->closable) ? 31 : (dwOpt->floatable || dwOpt->closable) ? 16 : 0;
		       if (verticalTitleBar)
		         r.adjust(0, offset, 0, 0);
		       else
		         r.adjust(0, 0, -offset, 0);
		       if (dwOpt->movable)
		         p->fillRect(r, cg.brush(QColorGroup::Highlight));
		     }
		     if (!dwOpt->title.isEmpty()) {
		       if (verticalTitleBar) {
		         QSize s = r.size();
		         s.transpose();
		         r.setSize(s);

		         p->save();
		         p->setRenderHint(QPainter::Antialiasing, false);
		         p->translate(r.left(), r.top() + r.width());
		         p->rotate(-90);
		         p->translate(-r.left(), -r.top());
		       }

		       const int indent = p->fontMetrics().descent();
		       drawItemText(p, r.adjusted(indent + 1, 1, -indent - 1, -1),
		                    (floating ? Qt::AlignLeft : Qt::AlignHCenter) | Qt::AlignVCenter
		                    | Qt::TextShowMnemonic, cg, true, dwOpt->title,
		                    (floating || dwOpt->movable) ? QPalette::HighlightedText : QPalette::WindowText);

		       if (verticalTitleBar)
		         p->restore();
		     }
		   }
		   break;

		default: {
		   QCleanlooksStyle::drawControl(control, option, p, widget);
		   break;
		}
	}
}

#if 0
void QuarticurveStyle::drawControlMask(ControlElement control,
				      QPainter *p,
				      const QWidget *widget,
				      const QRect &r,
				      const QStyleOption &data) const
{
    switch (control) {
		case CE_PushButton: {
			p->fillRect(r, Qt::color1);
			break;
		}

		default: {
			QCleanlooksStyle::drawControlMask(control, p, widget, r, data);
			break;
		}
    }
}
#endif

QRect QuarticurveStyle::subElementRect(SubElement subrect, const QStyleOption *option, const QWidget *widget) const
{
    QRect rect;

    switch (subrect) {
		case QStyle::SE_PushButtonFocusRect: {
			const QStyleOptionButton *buttonOption = qstyleoption_cast<const QStyleOptionButton *>(option);
			int dbw1 = 0, dbw2 = 0;
			if (buttonOption && ((buttonOption->features & QStyleOptionButton::DefaultButton) || (buttonOption->features & QStyleOptionButton::AutoDefaultButton))) {
				dbw1 = pixelMetric(PM_ButtonDefaultIndicator, option, widget);
				dbw2 = dbw1 * 2;
			}

			rect.setRect(option->rect.x()	  + 3 + dbw1,
					 option->rect.y()	  + 3 + dbw1,
					 option->rect.width()  - 6 - dbw2,
					 option->rect.height() - 6 - dbw2);
			break;
		}

		case QStyle::SE_CheckBoxIndicator: {
			int h = pixelMetric( PM_IndicatorHeight, option );
			rect.setRect(option->rect.x() + ( option->rect.height() - h ) / 2,
						 option->rect.y() + ( option->rect.height() - h ) / 2,
						 pixelMetric( PM_IndicatorWidth, option ), h );
			rect = visualRect(option->direction, option->rect, rect);
			break;
		}

		case QStyle::SE_RadioButtonIndicator: {
			int h = pixelMetric( PM_ExclusiveIndicatorHeight, option );
			rect.setRect( option->rect.x() + ( option->rect.height() - h ) / 2,
						  option->rect.y() + ( option->rect.height() - h ) / 2,
						  pixelMetric( PM_ExclusiveIndicatorWidth, option ), h );
			rect = visualRect(option->direction, option->rect, rect);
			break;
		}

		default: {
			rect = QCleanlooksStyle::subElementRect(subrect, option, widget);
		}
    }

    return rect;
}

void QuarticurveStyle::drawComplexControl(ComplexControl control,
					 const QStyleOptionComplex *option,
					 QPainter* p,
					 const QWidget* widget) const
{
    QColorGroup cg(adjustPalette(option->palette));
    QRect r(option->rect);
    QStyle::State flags(option->state);
    QStyle::SubControls controls(option->subControls), active(option->activeSubControls);
    const QuarticurveColorData *cdata = lookupData(cg);
    
    if (widget && widget == singleton->hoverWidget)
        flags |= QStyle::State_MouseOver;

    switch (control) {
#if 0
		case CC_ListView: {
			if ( controls & SC_ListView )
				QCleanlooksStyle::drawComplexControl( control, p, widget, r, cg, flags, controls, active, data );
			p->save();
			p->setRenderHint(QPainter::Antialiasing, false);
			if ( controls & ( SC_ListViewBranch | SC_ListViewExpand ) ) {
				if (data.isDefault())
					break;
				
				Q3ListViewItem *item = data.listViewItem(),
				*child = item->firstChild();

				int y = r.y();
				int c;
				int dotoffset = 0;
				Q3PointArray dotlines;
				if ( active == SC_All && controls == SC_ListViewExpand ) {
					c = 2;
					dotlines.resize(2);
					dotlines[0] = QPoint( r.right(), r.top() );
					dotlines[1] = QPoint( r.right(), r.bottom() );
				} else {
					int linetop = 0, linebot = 0;
					// each branch needs at most two lines, ie. four end points
					dotoffset = (item->itemPos() + item->height() - y) %2;
					dotlines.resize( item->childCount() * 4 );
					c = 0;
					
					// skip the stuff above the exposed rectangle
					while ( child && y + child->height() <= 0 ) {
						y += child->totalHeight();
						child = child->nextSibling();
					}
					
					int bx = r.width() / 2;
					
					// paint stuff in the magical area
					Q3ListView* v = item->listView();
					int lh;
					if ( !item->multiLinesEnabled() )
						lh = child ? child->height() : 0;
					else
						lh = p->fontMetrics().height() + 2 * v->itemMargin();
					lh = QMAX( lh, QApplication::globalStrut().height() );
					if ( lh % 2 > 0 )
						lh++;
					while ( child && y < r.height() ) {
						linebot = y + lh/2;
						if ( (child->isExpandable() || child->childCount()) &&
							 (child->height() > 0) ) {
							// needs a box
							p->setPen( cg.mid() );
							p->drawRect( bx-4, linebot-4, 9, 9 );
							// plus or minus
							p->setPen( cg.text() );
							p->drawLine( bx - 2, linebot, bx + 2, linebot );
							if ( !child->isOpen() )
								p->drawLine( bx, linebot - 2, bx, linebot + 2 );
							// dotlinery
							p->setPen( cg.mid() );
							dotlines[c++] = QPoint( bx, linetop );
							dotlines[c++] = QPoint( bx, linebot - 4 );
							dotlines[c++] = QPoint( bx + 5, linebot );
							dotlines[c++] = QPoint( r.width(), linebot );
							linetop = linebot + 5;
						} else {
							// just dotlinery
							dotlines[c++] = QPoint( bx+1, linebot );
							dotlines[c++] = QPoint( r.width(), linebot );
						}
						
						y += child->totalHeight();
						child = child->nextSibling();
					}
					
					if ( child ) // there's a child, so move linebot to edge of rectangle
						linebot = r.height();
					
					if ( linetop < linebot ) {
						dotlines[c++] = QPoint( bx, linetop );
						dotlines[c++] = QPoint( bx, linebot );
					}
				}
				p->setPen( cg.text() );

				static QBitmap *verticalLine = 0, *horizontalLine = 0;
				static Q3CleanupHandler<QBitmap> qlv_cleanup_bitmap;
				if ( !verticalLine ) {
					// make 128*1 and 1*128 bitmaps that can be used for
					// drawing the right sort of lines.
					verticalLine = new QBitmap( 1, 129, TRUE );
					horizontalLine = new QBitmap( 128, 1, TRUE );
					Q3PointArray a( 64 );
					QPainter p;
					p.begin( verticalLine );
					int i;
					for (i = 0; i < 64; i++)
						a.setPoint( i, 0, i*2+1 );
					p.setPen( Qt::color1 );
					p.drawPoints( a );
					p.end();
					QApplication::flushX();
					verticalLine->setMask( *verticalLine );
					p.begin( horizontalLine );
					for (i = 0; i < 64; i++)
						a.setPoint( i, i*2+1, 0 );
					p.setPen( Qt::color1 );
					p.drawPoints( a );
					p.end();
					QApplication::flushX();
					horizontalLine->setMask( *horizontalLine );
					qlv_cleanup_bitmap.add( &verticalLine );
					qlv_cleanup_bitmap.add( &horizontalLine );
				}
				
				int line; // index into dotlines
				for (line = 0; line < c; line += 2) {
					// assumptions here: lines are horizontal or vertical.
					// lines always start with the numerically lowest
					// coordinate.
					
					// point ... relevant coordinate of current point
					// end ..... same coordinate of the end of the current line
					// other ... the other coordinate of the current point/line
					if (dotlines[line].y() == dotlines[line+1].y() ) {
						int end = dotlines[line+1].x();
						int point = dotlines[line].x();
						int other = dotlines[line].y();
						while (point < end) {
							int i = 128;
							if ( i+point > end )
								i = end-point;
							p->drawPixmap( point, other, *horizontalLine,
										   0, 0, i, 1 );
							point += i;
						}
					} else {
						int end = dotlines[line+1].y();
						int point = dotlines[line].y();
						int other = dotlines[line].x();
						int pixmapoffset = ((point & 1) != dotoffset ) ? 1 : 0;
						while (point < end) {
							int i = 128;
							if ( i+point > end )
								i = end-point;
							p->drawPixmap( other, point, *verticalLine,
										   0, pixmapoffset, 1, i );
							point += i;
						}
					}
				}
			}
			p->restore();
			break;
		}
#endif
		case CC_ComboBox: {
			const QStyleOptionComboBox *comboboxOption = qstyleoption_cast<const QStyleOptionComboBox *>(option);
			QRect frame, arrow, field;

			frame=QStyle::visualRect(option->direction, option->rect, subControlRect(CC_ComboBox, option, SC_ComboBoxFrame, widget));
			arrow=QStyle::visualRect(option->direction, option->rect, subControlRect(CC_ComboBox, option, SC_ComboBoxArrow, widget));
			field=QStyle::visualRect(option->direction, option->rect, subControlRect(CC_ComboBox, option, SC_ComboBoxEditField, widget));


			if ((controls & SC_ComboBoxFrame) && frame.isValid()) {
				if (flags & QStyle::State_MouseOver)
					drawLightBevel(p, frame, cg, flags | QStyle::State_Raised, &cg.brush(QColorGroup::Midlight), true);
				else
					drawLightBevel(p, frame, cg, flags | QStyle::State_Raised, &cg.brush(QColorGroup::Button), true);
			}

			if ((controls & SC_ComboBoxArrow) && arrow.isValid()) {
				QStyleOption changedOptions(*option);
				changedOptions.state &= ~QStyle::State_MouseOver;
				changedOptions.rect = arrow;
				drawPrimitive(QStyle::PE_IndicatorArrowDown, &changedOptions, p, widget);
				p->save();
				p->setPen(cdata->shades[3]);
				p->setRenderHint(QPainter::Antialiasing, false);
				p->drawLine((arrow.left()+arrow.right())/2-2, 
							(arrow.top()+arrow.bottom())/2+5, 
							(arrow.left()+arrow.right())/2+2,
							(arrow.top()+arrow.bottom())/2+5);
				p->drawLine((arrow.left()+arrow.right())/2-2,
							(arrow.top()+arrow.bottom())/2+6,
							(arrow.left()+arrow.right())/2+2,
							(arrow.top()+arrow.bottom())/2+6);
				p->restore();
			}

			if ((controls & SC_ComboBoxEditField) && field.isValid()) {
				p->save();
				p->setPen(cdata->shades[4]);
				p->setRenderHint(QPainter::Antialiasing, false);
				if (comboboxOption && comboboxOption->editable) {
					field.addCoords(-1, -1, 1, 1);
					p->drawLine(field.right(), field.top() - 1,
								field.right(), field.bottom() + 1);
					p->setPen(cg.light());
					p->drawLine(field.right() + 1, field.top(),
								field.right() + 1, field.bottom() );
				} else {
					p->drawLine(field.right() + 1, field.top() - 2,
								field.right() + 1, field.bottom() + 2);
					p->setPen(cg.light());
					p->drawLine(field.right() + 2, field.top() - 1,
								field.right() + 2, field.bottom() + 1);
				}
				p->restore();

				if (flags & QStyle::State_HasFocus) {
					if (!comboboxOption || !comboboxOption->editable) {
						QRect fr=QStyle::visualRect(option->direction, option->rect, subElementRect(QStyle::SE_ComboBoxFocusRect, option, widget) );
                        fr.setRight(fr.right() - 3);
						QStyleOption changedOptions(*option);
						changedOptions.state |= QStyle::State_FocusAtBorder;
						changedOptions.rect = fr;
						drawPrimitive(QStyle::PE_FrameFocusRect, &changedOptions, p, widget);
					}

				}
#if 0 /* don't set the pen settings here (Kevin Kofler) */
				if (flags & QStyle::State_Enabled)
					p->setPen(cg.buttonText());
				else
					p->setPen(cg.mid());
#endif
			}

			break;
		}

		case CC_SpinBox: {
			const QStyleOptionSpinBox *spinboxOption = qstyleoption_cast<const QStyleOptionSpinBox *>(option);
			QRect frame, up, down;

			frame = subControlRect(CC_SpinBox, option, SC_SpinBoxFrame, widget);
			up = subControlRect(CC_SpinBox, option, SC_SpinBoxUp, widget);
			down = subControlRect(CC_SpinBox, option, SC_SpinBoxDown, widget);

			if ((controls & SC_SpinBoxFrame) && frame.isValid())
				drawTextRect(p, frame, cg, flags | QStyle::State_Sunken, &cg.brush(QColorGroup::Base));

			p->save();
			p->setPen(cdata->shades[5]);
			p->setRenderHint(QPainter::Antialiasing, false);
			p->drawLine(up.topLeft(), down.bottomLeft());
			p->drawLine(up.left(), up.bottom()+1, up.right(), up.bottom()+1);
			p->restore();

			if ((controls & SC_SpinBoxUp) && up.isValid()) {
				PrimitiveElement pe = QStyle::PE_IndicatorSpinUp;
				if (spinboxOption && spinboxOption->buttonSymbols == QAbstractSpinBox::PlusMinus)
				  pe = QStyle::PE_IndicatorSpinPlus;

				bool sunken=(active == SC_SpinBoxUp) && (flags & QStyle::State_Sunken);
				up.addCoords(1, 0, 0, 0);
				p->save();
				p->setRenderHint(QPainter::Antialiasing, false);
				p->fillRect(up, cg.brush(sunken?QColorGroup::Dark:(active == SC_SpinBoxUp)?QColorGroup::Midlight:QColorGroup::Button));
				if (sunken)
				  p->setPen(cdata->shades[2]);
				else
				  p->setPen(cg.light());
				p->drawLine(up.left(), up.top(), up.right(), up.top());
				p->drawLine(up.left(), up.top(), up.left(), up.bottom());
				if (sunken)
				  p->setPen(cg.light());
				else
				  p->setPen(cdata->shades[2]);
				p->drawLine(up.right(), up.top()+1, up.right(), up.bottom());
				p->drawLine(up.left()+1, up.bottom(), up.right(), up.bottom());
				p->restore();

				up.addCoords(1, 0, 0, 0);
				QStyleOption changedOptions(*option);
				changedOptions.state |= (sunken ?
							 QStyle::State_On | QStyle::State_Sunken : QStyle::State_Raised);
				if (!sunken) changedOptions.state &= ~QStyle::State_Sunken;
				changedOptions.rect = up;
				drawPrimitive(pe, &changedOptions, p, widget);
			}

			if ((controls & SC_SpinBoxDown) && down.isValid()) {
				PrimitiveElement pe = QStyle::PE_IndicatorSpinDown;
				if (spinboxOption && spinboxOption->buttonSymbols == QAbstractSpinBox::PlusMinus)
				  pe = QStyle::PE_IndicatorSpinMinus;

				bool sunken=(active == SC_SpinBoxDown) && (flags & QStyle::State_Sunken);
				down.addCoords(1, 0, 0, 0);
				p->save();
				p->setRenderHint(QPainter::Antialiasing, false);
				p->fillRect(down, cg.brush(sunken?QColorGroup::Dark:(active == SC_SpinBoxDown)?QColorGroup::Midlight:QColorGroup::Button));
				if (sunken)
				  p->setPen(cdata->shades[2]);
				else
				  p->setPen(cg.light());
				p->drawLine(down.left(), down.top(), down.right(), down.top());
				p->drawLine(down.left(), down.top(), down.left(), down.bottom());
				if (sunken)
				  p->setPen(cg.light());
				else
				  p->setPen(cdata->shades[2]);
				p->drawLine(down.right(), down.top()+1,
							down.right(), down.bottom());
				p->drawLine(down.left()+1, down.bottom(),
							down.right(), down.bottom());
				p->restore();

				down.addCoords(1, 0, 0, 0);
				QStyleOption changedOptions(*option);
				changedOptions.state |= (sunken ?
							 QStyle::State_On | QStyle::State_Sunken : QStyle::State_Raised);
				if (!sunken) changedOptions.state &= ~QStyle::State_Sunken;
				changedOptions.rect = down;
				drawPrimitive(pe, &changedOptions, p, widget);
			}

			break;
		}

		case CC_ScrollBar: {
			const QStyleOptionSlider *sliderOption = qstyleoption_cast<const QStyleOptionSlider *>(option);
			if (!sliderOption) { // oops, panic
				QCleanlooksStyle::drawComplexControl(control, option, p, widget);
				break;
			}
			QRect addline, subline, addpage, subpage, slider, first, last;
			bool maxedOut = (sliderOption->minimum == sliderOption->maximum);

			subline = subControlRect(control, option, SC_ScrollBarSubLine, widget);
			addline = subControlRect(control, option, SC_ScrollBarAddLine, widget);
			subpage = subControlRect(control, option, SC_ScrollBarSubPage, widget);
			addpage = subControlRect(control, option, SC_ScrollBarAddPage, widget);
			slider  = subControlRect(control, option, SC_ScrollBarSlider,  widget);
			first   = subControlRect(control, option, SC_ScrollBarFirst,   widget);
			last    = subControlRect(control, option, SC_ScrollBarLast,    widget);

			// Mouse position, for highlighting
			const QWidget *w = widget;
			if (!w) {
				if (p && p->device()->devType() == QInternal::Widget)
					w = (QWidget *) p->device();
			}
			QPoint mp;
			if (w)
				mp = w->mapFromGlobal(QCursor::pos());
			if ((controls & SC_ScrollBarSubLine) && subline.isValid()) 
			{
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.state = QStyle::State(((w && subline.contains(mp)) ? QStyle::State_MouseOver : 0) | QStyle::State_Enabled | ((active == SC_ScrollBarSubLine) ? QStyle::State_Sunken : QStyle::State_Default) | ((sliderOption->orientation == Qt::Horizontal) ? QStyle::State_Horizontal : 0));
				changedOptions.rect = subline;
				drawControl(QStyle::CE_ScrollBarSubLine, &changedOptions, p, widget);
			}
			if ((controls & SC_ScrollBarAddLine) && addline.isValid()) 
			{
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.state = QStyle::State(((w && addline.contains(mp)) ? QStyle::State_MouseOver : 0) | QStyle::State_Enabled | ((active == SC_ScrollBarAddLine) ? QStyle::State_Sunken : QStyle::State_Default) | ((sliderOption->orientation == Qt::Horizontal) ? QStyle::State_Horizontal : 0));
				changedOptions.rect = addline;
				drawControl(QStyle::CE_ScrollBarAddLine, &changedOptions, p, widget);
			}
			if ((controls & SC_ScrollBarSubPage) && subpage.isValid())
			{
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.state = QStyle::State(QStyle::State_Enabled | ((active == SC_ScrollBarSubPage) ? QStyle::State_Sunken : QStyle::State_Default) | ((sliderOption->orientation == Qt::Horizontal) ? QStyle::State_Horizontal : 0));
				changedOptions.rect = subpage;
				drawControl(QStyle::CE_ScrollBarSubPage, &changedOptions, p, widget);
			}
			if ((controls & SC_ScrollBarAddPage) && addpage.isValid())
			{
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.state = QStyle::State(((maxedOut) ? QStyle::State_Default : QStyle::State_Enabled) | ((active == SC_ScrollBarAddPage) ? QStyle::State_Sunken : QStyle::State_Default) | ((sliderOption->orientation == Qt::Horizontal) ? QStyle::State_Horizontal : 0));
				changedOptions.rect = addpage;
				drawControl(QStyle::CE_ScrollBarAddPage, &changedOptions, p, widget);
			}
			if ((controls & SC_ScrollBarFirst) && first.isValid())
			{
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.state = QStyle::State(QStyle::State_Enabled | ((active == SC_ScrollBarFirst) ? QStyle::State_Sunken : QStyle::State_Default) | ((sliderOption->orientation == Qt::Horizontal) ? QStyle::State_Horizontal : 0));
				changedOptions.rect = first;
				drawControl(QStyle::CE_ScrollBarFirst, &changedOptions, p, widget);
			}
			if ((controls & SC_ScrollBarLast) && last.isValid())
			{
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.state = QStyle::State(QStyle::State_Enabled | ((active == SC_ScrollBarLast) ? QStyle::State_Sunken : QStyle::State_Default) | ((sliderOption->orientation == Qt::Horizontal) ? QStyle::State_Horizontal : 0));
				changedOptions.rect = last;
				drawControl(QStyle::CE_ScrollBarLast, &changedOptions, p, widget);
			}
			if ((controls & SC_ScrollBarSlider) && slider.isValid()) {
			{
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.state = QStyle::State(((w && slider.contains(mp)) ? QStyle::State_MouseOver : 0) | QStyle::State_Enabled | ((active == SC_ScrollBarSlider) ? QStyle::State_Sunken : QStyle::State_Default) | ((sliderOption->orientation == Qt::Horizontal) ? QStyle::State_Horizontal : 0));
				changedOptions.rect = slider;
				drawControl(QStyle::QStyle::CE_ScrollBarSlider, &changedOptions, p, widget);
			}

              /*
			  // ### perhaps this should not be able to accept focus if maxedOut?
			  if (scrollbar->hasFocus()) {
				QRect fr(slider.x() + 2, slider.y() + 2, slider.width() - 5, slider.height() - 5);
				drawPrimitive(QStyle::PE_FrameFocusRect, p, fr, cg, QStyle::State_Default);
			  }
              */
			}
			break;
		}

		case CC_Slider: {
			const QStyleOptionSlider *sliderOption = qstyleoption_cast<const QStyleOptionSlider *>(option);
			if (!sliderOption) { // oops, panic
				QCleanlooksStyle::drawComplexControl(control, option, p, widget);
				break;
			}
			QRect groove = subControlRect(CC_Slider, option, SC_SliderGroove, widget),
			handle = subControlRect(CC_Slider, option, SC_SliderHandle, widget);

			if ((controls & SC_SliderGroove) && groove.isValid()) {
			  if (flags & QStyle::State_HasFocus) {
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.rect = groove;
				drawPrimitive(QStyle::PE_FrameFocusRect, &changedOptions, p, widget);
			  }

			  if (sliderOption->orientation == Qt::Horizontal) {
				int dh = (groove.height() - 5) / 2;
				groove.addCoords(0, dh, 0, -dh);
				handle.addCoords(0, 1, 0, -1);
			  } else {
				int dw = (groove.width() - 5) / 2;
				groove.addCoords(dw, 0, -dw, 0);
				handle.addCoords(1, 0, -1, 0);
			  }

			  p->save();
			  p->setPen(cdata->shades[5]);
			  p->setBrush(cg.mid());
			  p->setRenderHint(QPainter::Antialiasing, false);
			  p->drawRect(groove);
			  p->setPen(cdata->shades[4]);
			  p->drawLine(groove.left()+1, groove.top()+1,
						  groove.left()+1, groove.bottom()-1);
			  p->drawLine(groove.left()+1, groove.top()+1,
						  groove.right()-1, groove.top()+1);
			  p->restore();
			}

			if ((controls & SC_SliderHandle) && handle.isValid()) {
			  p->save();
			  p->setPen(cdata->shades[6]);
			  p->setRenderHint(QPainter::Antialiasing, false);
			  p->drawLine(handle.x() + 2, handle.y(),
						  handle.right() - 2, handle.y());
			  p->drawLine(handle.x(), handle.y() + 2,
						  handle.x(), handle.bottom() - 2);
			  p->drawLine(handle.right(), handle.y() + 2,
						  handle.right(), handle.bottom() - 2);
			  p->drawLine(handle.x() + 2, handle.bottom(),
						  handle.right() - 2, handle.bottom());
			  p->drawPoint(handle.x() + 1, handle.y() + 1);
			  p->drawPoint(handle.right() - 1, handle.y() + 1);
			  p->drawPoint(handle.right() - 1, handle.bottom() - 1);
			  p->drawPoint (handle.x() + 1, handle.bottom() - 1);
			  

			  p->setPen(cdata->shades[2]);
			  p->drawLine(handle.x() + 2, handle.bottom() - 1,
						  handle.right() - 2, handle.bottom() - 1);
			  p->drawLine(handle.right() - 1, handle.top() + 2,
						  handle.right() - 1, handle.bottom() - 2);
			  p->drawPoint (handle.x() + 1, handle.y());
			  p->drawPoint (handle.right() - 1, handle.y());
			  p->drawPoint (handle.x(), handle.y() + 1);
			  p->drawPoint (handle.right(), handle.y() + 1);
			  p->drawPoint (handle.x(), handle.bottom() - 1);
			  p->drawPoint (handle.right(), handle.bottom() - 1);
			  p->drawPoint (handle.x() + 1, handle.bottom() );
			  p->drawPoint (handle.right() - 1, handle.bottom());
			  
			  p->setPen(Qt::white);
			  p->drawLine (handle.x() + 2, handle.y() + 1,
						   handle.right() - 2, handle.y() + 1);
			  p->drawLine (handle.x() + 1, handle.y() + 2,
						   handle.x() + 1, handle.bottom() - 2);


			  p->setBrush(cg.button());
			  QRect fillr (handle);
			  fillr.addCoords (2, 2, -2, -2);
			  p->fillRect (fillr, cg.brush(QColorGroup::Button));


			  if (sliderOption->orientation == Qt::Horizontal) {
				int x1 = handle.x() + handle.width() / 2 - 8;
				int y1 = handle.y() + (handle.height() - 6) / 2;
				p->setPen(cdata->shades[5]);
				p->drawLine(x1 + 5, y1,
							x1, y1 + 5);
				p->drawLine(x1 + 5 + 5, y1,
							x1 + 5, y1 + 5);
				p->drawLine(x1 + 5 + 5*2, y1,
							x1 + 5*2, y1 + 5);
				
				p->setPen(Qt::white);
				p->drawLine(x1 + 5, y1 + 1,
							x1 + 1, y1 + 5);
				p->drawLine(x1 + 5 + 5, y1 + 2,
							x1 + 5 + 1, y1 + 5);
				p->drawLine(x1 + 5 + 5*2, y1 + 1,
							x1 + 5*2 + 1, y1 + 5);
			  } else {
				int x1 = handle.x() + (handle.width() - 6) / 2;
				int y1 = handle.y() + handle.height() / 2 - 8;
				p->setPen(cdata->shades[5]);
				p->drawLine(x1 + 5, y1,
							x1, y1 + 5);
				p->drawLine(x1 + 5, y1 + 5,
							x1, y1 + 5 + 5);
				p->drawLine(x1 + 5, y1 + 5*2,
							x1, y1 + 5 + 5*2);
				p->setPen(Qt::white);
				p->drawLine(x1 + 5, y1 + 1,
							x1 + 1, y1 + 5);
				p->drawLine(x1 + 5, y1 + 5 + 1,
							x1 + 1, y1 + 5 + 5);
				p->drawLine(x1 + 5, y1 + 5*2 + 1,
							x1 + 1, y1 + 5 + 5*2);
			  }
			  p->restore();
			}

			if (controls & SC_SliderTickmarks) {
				QStyleOptionSlider changedOptions(*sliderOption);
				changedOptions.subControls = SC_SliderTickmarks;
				QCleanlooksStyle::drawComplexControl(control, &changedOptions, p, widget);
			}
			break;
		}

		default: {
			QCleanlooksStyle::drawComplexControl(control, option, p, widget);
			break;
		}
    }
}

QRect QuarticurveStyle::subControlRect( ComplexControl control,
					       const QStyleOptionComplex *option,
					       SubControl sc,
					       const QWidget *widget ) const
{
    QRect ret;

    switch (control) {
		case CC_ScrollBar: {
			const QStyleOptionSlider *sliderOption = qstyleoption_cast<const QStyleOptionSlider *>(option);
			if (!sliderOption) { // oops, panic
				ret = QCleanlooksStyle::subControlRect(control, option, sc, widget);
				break;
			}
			int sbextent = pixelMetric(PM_ScrollBarExtent, option, widget);
			int maxlen = ((sliderOption->orientation == Qt::Horizontal) ?
					  sliderOption->rect.width() : sliderOption->rect.height()) - (sbextent * 3);
			int sliderlen;

			// calculate slider length
			if (sliderOption->maximum != sliderOption->minimum) {
				uint range = sliderOption->maximum - sliderOption->minimum;
				sliderlen = (sliderOption->pageStep * maxlen) /
				(range + sliderOption->pageStep);

				int slidermin = pixelMetric( PM_ScrollBarSliderMin, option, widget );
				if ( sliderlen < slidermin || range > INT_MAX / 2 )
					sliderlen = slidermin;
				if ( sliderlen > maxlen )
					sliderlen = maxlen;
			} else
				sliderlen = maxlen;

			int sliderstart = sbextent + sliderPositionFromValue(sliderOption->minimum,
			                                                     sliderOption->maximum,
			                                                     sliderOption->sliderPosition,
			                                                     maxlen + sbextent - sliderlen,
			                                                     sliderOption->upsideDown);

			switch (sc) {
				case SC_ScrollBarSubLine: {
					// top/left button
					ret.setRect(0, 0, sbextent, sbextent);
					break;
				}

				case SC_ScrollBarAddLine: {
					// bottom/right button
					if (sliderOption->orientation == Qt::Horizontal)
						ret.setRect(sliderOption->rect.width() - sbextent, 0, sbextent, sbextent);
					else
						ret.setRect(0, sliderOption->rect.height() - sbextent, sbextent, sbextent);
					break;
				}

				case SC_ScrollBarSubPage: {
					// between top/left button and slider
					if (sliderOption->orientation == Qt::Horizontal)
						ret.setRect(sbextent, 0, sliderstart - sbextent, sbextent);
					else
						ret.setRect(0, sbextent, sbextent, sliderstart - sbextent);
					break;
				}

				case SC_ScrollBarAddPage: {
					// between bottom/right button and slider
					if (sliderOption->orientation == Qt::Horizontal)
						ret.setRect(sliderstart + sliderlen, 0,
									maxlen - sliderstart - sliderlen + sbextent * 2, sbextent);
					else
						ret.setRect(0, sliderstart + sliderlen,
									sbextent, maxlen - sliderstart - sliderlen + sbextent * 2);
					break;
				}

				case SC_ScrollBarGroove: {
					if (sliderOption->orientation == Qt::Horizontal)
						ret.setRect(sbextent, 0, sliderOption->rect.width() - sbextent * 2,
									sliderOption->rect.height());
					else
						ret.setRect(0, sbextent, sliderOption->rect.width(),
									sliderOption->rect.height() - sbextent * 2);
					break;
				}

				case SC_ScrollBarSlider: {
					if (sliderOption->orientation == Qt::Horizontal)
                        // here

						ret.setRect(sliderstart, 0, sliderlen, sbextent);
					else
						ret.setRect(0, sliderstart, sbextent, sliderlen);
					break;
				}

				default: {
					break;
				}
			}

			break;
		}

		case CC_SpinBox: {
			int fw = pixelMetric( PM_SpinBoxFrameWidth, option, widget);
				QRect r(0, 0, 0, 0);
				if (option)
					r = option->rect;
				else if (widget)
					r = widget->rect();
				QSize bs;
				bs.setHeight( r.height()/2 - fw );
				if ( bs.height() < 8 )
					bs.setHeight( 8 );
				bs.setWidth( bs.height() * 8 / 6 ); 
				bs = bs.expandedTo( QApplication::globalStrut() );
				int y = fw;
				int x, lx, rx;
				x = r.width() - y - bs.width() + 1;
				lx = fw;
				rx = x - fw;
				switch ( sc ) {
					case SC_SpinBoxUp: {
						ret.setRect(x, y-1, bs.width(), bs.height()+1);
						break;
					}
					case SC_SpinBoxDown: {
						ret.setRect(x, y + bs.height()+1, bs.width(), bs.height()+1);
						break;
					}
#if 0 /* apparently no longer needed in Qt 4 (Kevin Kofler) */
					case SC_SpinBoxButtonField: {
						ret.setRect(x, y, bs.width()+1, r.height() - 2*fw);
						break;
					}
#endif
					case SC_SpinBoxEditField: {
						ret.setRect(lx, fw, rx, r.height() - 2*fw);
						break;
					}
					case SC_SpinBoxFrame: {
						ret = r;
					}
					default: {
						break;
					}
				}

			break;
		}

        case CC_ComboBox : {
            /* bypass QCleanlooksStyle here (Kevin Kofler) */
            ret = QWindowsStyle::subControlRect(control, option, sc, widget);
            switch (sc) {
                case SC_ComboBoxFrame: {
                    break;
                }
                case SC_ComboBoxArrow: {
                    ret.setTop(ret.top() - 2);
                    ret.setLeft(ret.left() - 1);
                    break;
                }
                case SC_ComboBoxEditField: {
                    ret.setRight(ret.right() - 2);
                    break;
                }
			    default: {
				    break;
			    }
            }
            
            /*
            int h = ret.height(), w = ret.width();
            if (h < 25)
                h = 25;
            if (w < 20)
                w = 20;
            ret.setHeight(h);
            ret.setWidth(w);
            */
            break;
        }

		default: {
			ret = QCleanlooksStyle::subControlRect(control, option, sc, widget);
			break;
		}
    }

    return ret;
}

int QuarticurveStyle::pixelMetric( PixelMetric metric, const QStyleOption *option, const QWidget *widget ) const
{
    int ret;

    switch (metric) {
		case PM_ButtonMargin: {
			ret = 10;
			break;
		}

		case PM_ButtonShiftHorizontal:
		case PM_ButtonShiftVertical: {
			ret = 0;
			break;
		}

		case PM_ButtonDefaultIndicator: {
			ret = 0;
			break;
		}

		case PM_ListViewIconSize:
		case PM_ButtonIconSize: {
			ret = 16;
			break;
		}

		case PM_MessageBoxIconSize: {
			ret = 32;
			break;
		}

		case PM_DefaultFrameWidth: {
			if (widget && widget->inherits("QMenu")) {
				ret = 3;
			} else  if (widget && widget->inherits("Q3WidgetStack")) {
				ret = 2;
			} else {
                ret = 1;
            }
			break;
		}

		case PM_IndicatorWidth:
		case PM_IndicatorHeight:
		case PM_ExclusiveIndicatorWidth:
		case PM_ExclusiveIndicatorHeight: {
			ret = 13;
			break;
		}

        case PM_MenuVMargin: {
            ret = 1;
            break;
        }

		case PM_TabBarTabOverlap: {
			ret = 1;
			break;
		}

        case PM_TabBarTabHSpace: {
            ret = 10;
            break;
        }

        case PM_TabBarTabVSpace: {
            ret = 10;
            break;
        }

        case PM_TabBarTabShiftHorizontal: {
            ret = 0;
            break;
        }

		case PM_ScrollBarExtent: {
			ret = 15;
			break;
		}

		case PM_MenuBarPanelWidth: {
			ret = 1;
			break;
		}

		case PM_ProgressBarChunkWidth: {
			ret = 2;
			break;
		}

		case PM_DockWidgetSeparatorExtent: {
			ret = 4;
			break;
		}

		case PM_DockWidgetHandleExtent: {
			ret = 10;
			break;
		}

		case PM_SplitterWidth: {
			ret = 6;
			break;
		}

		case PM_ScrollBarSliderMin: {
			ret=31;
			break;
		}

		case PM_SliderControlThickness: {
			ret = QCleanlooksStyle::pixelMetric( metric, option, widget );
			break;
		}

		case PM_SliderLength: {
			ret=31;
			const QStyleOptionSlider *sliderOption = qstyleoption_cast<const QStyleOptionSlider *>(option);
			if (sliderOption) {
				if (sliderOption->orientation == Qt::Horizontal) {
					if (option->rect.width()<ret)
						ret = option->rect.width();
				} else {
					if (option->rect.height()<ret)
						ret = option->rect.height();
				}
			}
			break;
		}

		case PM_MaximumDragDistance: {
			ret = -1;
			break;
		}

		default: {
			ret = QCleanlooksStyle::pixelMetric(metric, option, widget);
			break;
		}
    }

    return ret;
}

QSize QuarticurveStyle::sizeFromContents( ContentsType contents,
					 const QStyleOption *option,
					 const QSize &contentsSize,
					 const QWidget *widget ) const
{
    QSize ret = QCleanlooksStyle::sizeFromContents( contents, option, contentsSize, widget );

    switch (contents) {
    
		case CT_PushButton: {
			const QStyleOptionButton *buttonOption = qstyleoption_cast<const QStyleOptionButton *>(option);
			int w = ret.width(), h = ret.height();

			// only expand the button if we are displaying text...
			if (!buttonOption || buttonOption->icon.isNull() || !buttonOption->text.isEmpty()) {
				if ( w < 85 )
					w = 85;
				if ( h < 30 )
					h = 30;
			}

			ret = QSize( w, h );
			break;
		}

		case CT_MenuItem: {
			const QStyleOptionMenuItem *mi = qstyleoption_cast<const QStyleOptionMenuItem *>(option);
			if ( !mi )
				break;
			int maxpmw = mi->maxIconWidth;
			int w = contentsSize.width(), h = contentsSize.height();

			/* if (mi->custom()) {
				w = mi->custom()->sizeHint().width();
				h = mi->custom()->sizeHint().height();
				if (! mi->custom()->fullSpan() && h < 22)
				h = 22;
			} else if (mi->widget()) {
			} else */ if (mi->menuItemType == QStyleOptionMenuItem::Separator) {
				w = 10;
				h = 12;
			} else {
				// check is at least 16x16
				if (h < 16) {
					h = 16;
				}

				/* if (mi->pixmap()) {
					h = QMAX(h, mi->pixmap()->height() + 6);
				} else */ if (! mi->text.isNull() && widget) {
					h = QMAX(h, widget->fontMetrics().height() + 8);
				}
				if (! mi->icon.isNull()) {
					h = QMAX(h, mi->icon.pixmap(16, QIcon::Normal).height() + 6);
				}
			}

			// check is at least 16x16
			maxpmw = QMAX(maxpmw, 16);
			w += maxpmw + 16;

			if (!mi->text.isNull() && mi->text.find('\t') >= 0)
				w += 8;

			ret = QSize(w, h);
			break;
		}

		case CT_ToolButton: {
			int w = ret.width(), h = ret.height();
			h = (h < 32 ? 32 : h);
			w = (w < 32 ? 32 : w);
			ret = QSize(w, h);
			break;
		}

		case CT_ComboBox: {
			int w = ret.width(), h = ret.height();
			if (h < 27)
				h = 27;
			ret = QSize(w, h);
			break;
		}

        case CT_SpinBox: {
            int w = ret.width(), h = ret.height();
            if (h < 25)
                h = 25;

            ret = QSize(w, h);
            break; 

        }

		default: {
			break;
		}
    }

    return ret;
}

int QuarticurveStyle::styleHint( StyleHint stylehint,
				const QStyleOption *option,
				const QWidget *widget,
				QStyleHintReturn* returnData ) const
{
    int ret;

    switch (stylehint) {
		case SH_EtchDisabledText:
		case SH_Slider_SnapToValue:
		case SH_PrintDialog_RightAlignButtons:
		case SH_FontDialog_SelectAssociatedText:
		case SH_MenuBar_AltKeyNavigation:
		case SH_MenuBar_MouseTracking:
		case SH_Menu_MouseTracking:
		case SH_Menu_SpaceActivatesItem:
		case SH_ComboBox_ListMouseTracking: {
            ret = 1;
            break;
        }
// it needs more works for this feature
#if 0
        case SH_ComboBox_Popup: {
            const QComboBox *combobox = (const QComboBox *) widget;
            if (combobox->editable()) 
                ret = 0;
            else
                ret = 1;
            break;
        } 
#else
		case SH_ComboBox_Popup: {
			ret = 0;
			break;
		}
#endif
		case SH_ScrollBar_MiddleClickAbsolutePosition: {
			ret = 1;
			break;
		}

		case SH_MainWindow_SpaceBelowMenuBar: {
			ret = 0;
			break;
		}

		case SH_Menu_AllowActiveAndDisabled: {
			ret = 0;
			break;
		}

		case SH_DialogButtonLayout: {
			ret = QDialogButtonBox::KdeLayout;
			break;
		}

#if 0
        case SH_GUIStyle: {
            ret = Qt::GtkStyle;
            break;
        }
#endif

		default: {
			ret = QCleanlooksStyle::styleHint(stylehint, option, widget, returnData);
			break;
		}
    }

    return ret;
}

void QuarticurveStyle::drawGradient(QPainter *p, QRect const &rect,
		const QColorGroup &cg,
		double shade1, double shade2,
		bool horiz) const
{
	QColor c, c1, c2;
	int r, g, b;
	int c2r, c2g, c2b;
	int dr, dg, db, size;
	int start, end, left, right, top, bottom;

	left = rect.left();
	top = rect.top();
	bottom = rect.bottom();
	right = rect.right();

	start = horiz ? left : top;
	end = horiz ? right : bottom;

	if (end == start)
		return;

	shade (cg.highlight(), c1, shade1);
	shade (cg.highlight(), c2, shade2);

	c1.rgb(&r, &g, &b);
	c2.rgb(&c2r, &c2g, &c2b);

	size = end - start;
	dr = (c2r - r) / size;
	dg = (c2g - g) / size;
	db = (c2b - b) / size;

	p->save();
	p->setRenderHint(QPainter::Antialiasing, false);
	for (int i = start; i <= end; i++) {
		c.setRgb (r, g, b);
		p->setPen(c);
		if (horiz)
			p->drawLine(i, top, i, bottom);
		else
			p->drawLine(left, i, right, i);

		r += dr;
		g += dg;
		b += db;
	}
	p->restore();
}

void QuarticurveStyle::drawGradientBox(QPainter *p, QRect const &r,
                                     const QColorGroup &cg,
                                     const QuarticurveColorData *cdata,
                                     bool horiz,
                                     double shade1, double shade2) const
{
	QRect grad(r.left()+2, r.top()+2, r.width()-3, r.height()-3);
	drawGradient(p, grad, cg, shade1, shade2, horiz);

// 3d border effect...
	p->save();
	p->setPen(cdata->spots[2]);
	p->setBrush(Qt::NoBrush);
	p->setRenderHint(QPainter::Antialiasing, false);
	QRect r2(r);
	r2.adjust(0, 0, -1, -1);
	p->drawRect(r2);

//	We draw the bottom and right lines first ...
	p->setPen(cdata->spots[1]);
	p->drawLine(r.left()+1, r.bottom()-1, r.right()-1, r.bottom()-1);
	p->drawLine(r.right()-1, r.top()+1, r.right()-1, r.bottom()-1);

//	Because the lighter lines should overlap them on the corner pixels
	p->setPen(cdata->spots[0]);
	p->drawLine(r.left()+1, r.top()+1, r.right()-1, r.top()+1);
	p->drawLine(r.left()+1, r.top()+1, r.left()+1, r.bottom()-1);
	p->restore();
}

void QuarticurveStyle::drawItem( QPainter *p, const QRect &r,
                       int flags, const QPalette &g, bool enabled,
                       const QPixmap *pixmap, const QString& text,
                       QPalette::ColorRole textRole ) const
{
    if (pixmap) {
        QStyleOption option;
        option.rect=r;
        option.palette=g;
        QPixmap pm = generatedIconPixmap(enabled?QIcon::Normal:QIcon::Disabled, *pixmap, &option);
        drawItemPixmap(p, r, flags, pm);
    } else if (!text.isNull()) {
        drawItemText(p, r, flags, g, enabled, text, textRole);
    }
}

QPixmap QuarticurveStyle::standardPixmap(StandardPixmap standardPixmap,
                                         const QStyleOption *opt,
                                         const QWidget *widget) const
{
    if (standardPixmap ==  SP_TitleBarNormalButton
        || standardPixmap ==  SP_TitleBarMinButton
        || standardPixmap ==  SP_TitleBarCloseButton
        || standardPixmap ==  SP_TitleBarNormalButton
        || standardPixmap ==  SP_DockWidgetCloseButton)
        return QWindowsStyle::standardPixmap(standardPixmap, opt, widget);
    // HACK ALERT: We're violating the contract of setDesktopSettingsAware here, but it works...
    QApplication::setDesktopSettingsAware(true);
    return QCleanlooksStyle::standardPixmap(standardPixmap, opt, widget);
}

QIcon QuarticurveStyle::standardIconImplementation(StandardPixmap standardPixmap,
                                                   const QStyleOption *opt,
                                                   const QWidget *widget) const
{
    // HACK ALERT: We're violating the contract of setDesktopSettingsAware here, but it works...
    QApplication::setDesktopSettingsAware(true);
    return QCleanlooksStyle::standardIconImplementation(standardPixmap, opt, widget);
}

// vim:ts=4:sw=4
